/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"
#include "Poly.hh"
#include "Scalar_Prod.hh"
#include "Poly_templ.hh"
#include "Poly_min.hh"
#include "Rational.hh"

#include <algorithm>
#include <functional>

namespace pplite {

namespace detail {

bool
check_sat(const Sat& sat_g, const Cons& cons, const Gens& gens) {
  for (auto i = num_rows(cons); i-- > 0; ) {
    const auto& sat_i = sat_g[i];
    for (auto j = num_rows(gens); j-- > 0; ) {
      const bool saturated = !sat_i[j];
      const int s = Scalar_Prod::sign(cons[i], gens[j]);
      if (s < 0) return false;
      if (saturated && s > 0) return false;
      if (!saturated && s == 0) return false;
    }
  }
  return true;
}

} // namespace detail

void
Poly_Impl::clear_all() {
  // Clears everything but topol, dim and status.
  cs.clear();
  gs.clear();
  sat_c.clear();
  sat_g.clear();
  cs_pending.clear();
  gs_pending.clear();
}

void
Poly_Impl::set_empty() {
  clear_all();
  // Add `1 == 0'.
  cs.sing_rows.push_back(Con::zero_dim_false());
  set_status(Status::EMPTY);
}

void
Poly_Impl::set_universe() {
  clear_all();
  detail::set_universe(dim, cs, gs, sat_c);
  sat_g.transpose_assign(sat_c);
  set_status(Status::MINIMIZED);
}

void
Poly_Impl::set_topology(Topol t) {
  // Topological closure is assumed to hold.
  assert(topol == Topol::CLOSED || t == Topol::NNC
         || is_topologically_closed());
  topol = t;
}

void
Poly_Impl::reinit_with_gens(dim_type d, Topol t, Sys<Gens>&& gens) {
  assert(&gens != &gs && &gens != &gs_pending);
  dim = d;
  topol = t;
  clear_all();
  gs_pending = std::move(gens);
  detail::init_dd(dim, gs, gs_pending, cs, sat_g);
  sat_c.transpose_assign(sat_g);
  set_status(gs_pending.empty() ? Status::MINIMIZED : Status::PENDING);
  assert(check_inv());
}

Poly_Impl::Poly_Impl(dim_type d, Spec_Elem s, Topol t)
  : topol(t), dim(d) {
  if (s == Spec_Elem::EMPTY)
    set_empty();
  else
    set_universe();
}

bool
Poly_Impl::is_empty() const {
  if (has_valid_gens()) {
    // (NS_g = \empty && SK_g has no point) --> marked_empty()
    assert(!(gs.ns_rows.empty() && !has_point(gs.sk_rows))
           || marked_empty());
    return marked_empty();
  }
  minimize();
  return marked_empty();
}

bool
Poly_Impl::is_universe() const {
  if (marked_empty())
    return false;
  if (space_dim() == 0)
    return true;
  if (has_valid_cons()) {
    auto& pend_sing = cs_pending.sing_rows;
    auto& pend_sk = cs_pending.sk_rows;
    auto& pend_ns = cs_pending.ns_rows;
    const auto is_tauto = std::mem_fn(&Con::is_tautological);
    return cs.sing_rows.empty() && cs.ns_rows.empty()
      && std::all_of(cs.sk_rows.begin(), cs.sk_rows.end(), is_tauto)
      && pend_sing.empty() && pend_ns.empty()
      && std::all_of(pend_sk.begin(), pend_sk.end(), is_tauto);
  }
  minimize();
  return !marked_empty() && (num_lines() == space_dim());
}

bool
Poly_Impl::is_bounded() const {
  if (marked_empty() || space_dim() == 0)
    return true;
  if (has_valid_gens())
    return gs.sing_rows.empty()
      && !has_line_or_ray(gs.sk_rows)
      && gs_pending.sing_rows.empty()
      && !has_line_or_ray(gs_pending.sk_rows);
  minimize();
  return gs.sing_rows.empty() && !has_line_or_ray(gs.sk_rows);
}

bool
Poly_Impl::is_topologically_closed() const {
  if (is_necessarily_closed())
    return true;
  if (marked_empty() || space_dim() == 0)
    return true;
  minimize();
  return gs.ns_rows.empty() && !has_closure_point(gs.sk_rows);
}

dim_type
Poly_Impl::affine_dim() const {
  minimize();
  return marked_empty() ? 0 : space_dim() - cs.sing_rows.size();
}

dim_type
Poly_Impl::num_min_cons() const {
  minimize();
  auto num_cons = cs.sing_rows.size() + cs.sk_rows.size() + cs.ns_rows.size();
  bool has_pos = std::any_of(cs.sk_rows.cbegin(), cs.sk_rows.cend(),
                             std::mem_fn(&Con::is_tautological));
  if (has_pos)
    return num_cons - 1;
  bool has_efc = detail::has_efc(cs.ns_rows, sat_g, gs.sk_rows);
  if (has_efc)
    return num_cons - 1;
  // No positivity nor efc.
  return num_cons;
}

dim_type
Poly_Impl::num_min_gens() const {
  minimize();
  return gs.sing_rows.size() + gs.sk_rows.size() + gs.ns_rows.size();
}

namespace detail {

Poly_Con_Rel
relation_with(const Poly_Impl::Gens_Proxy& gs, const Con& c) {
  Poly_Con_Rel res = Poly_Con_Rel::saturates();

  switch (c.type()) {

  case Con::EQUALITY:
    {
      res = res && Poly_Con_Rel::is_included();
      // The following variable will hold the scalar product sign
      // of either the first point or the first non-saturating ray.
      // It starts equal to 2, meaning that we haven't found any yet.
      int first_point_or_nonsat_ray_sign = 2;
      for (const auto& g : gs) {
        const int sp_sign = Scalar_Prod::sign(c, g);
        if (sp_sign == 0) {
          if (g.is_point()) {
            if (first_point_or_nonsat_ray_sign == 2)
              first_point_or_nonsat_ray_sign = 0;
            else if (first_point_or_nonsat_ray_sign != 0)
              return Poly_Con_Rel::strictly_intersects();
          }
        } else {
          assert(sp_sign != 0);
          switch (g.type()) {
          case Gen::LINE:
            return Poly_Con_Rel::strictly_intersects();
          case Gen::RAY:
            if (first_point_or_nonsat_ray_sign == 2) {
              first_point_or_nonsat_ray_sign = sp_sign;
              res = Poly_Con_Rel::is_disjoint();
            }
            else if (sp_sign != first_point_or_nonsat_ray_sign)
              return Poly_Con_Rel::strictly_intersects();
            break;
          case Gen::POINT:
          case Gen::CLOSURE_POINT:
            // A non-saturating closure point is treated as a point.
            if (first_point_or_nonsat_ray_sign == 2) {
              first_point_or_nonsat_ray_sign = sp_sign;
              res = Poly_Con_Rel::is_disjoint();
            }
            else if (sp_sign != first_point_or_nonsat_ray_sign)
              return Poly_Con_Rel::strictly_intersects();
            break;
          } // switch (g.type())
        }
      }
    }
    break;

  case Con::NONSTRICT_INEQUALITY:
    {
      res = res && Poly_Con_Rel::is_included();
      bool no_point_or_nonsat_ray = true;
      for (const auto& g : gs) {
        const int sp_sign = Scalar_Prod::sign(c, g);
        if (sp_sign == 0) {
          if (g.is_point()) {
            if (no_point_or_nonsat_ray)
              no_point_or_nonsat_ray = false;
            else if (res == Poly_Con_Rel::is_disjoint())
              return Poly_Con_Rel::strictly_intersects();
          }
        } else {
          assert(sp_sign != 0);
          switch (g.type()) {
          case Gen::LINE:
            return Poly_Con_Rel::strictly_intersects();
          case Gen::RAY:
            if (no_point_or_nonsat_ray) {
              no_point_or_nonsat_ray = false;
              res = (sp_sign > 0)
                ? Poly_Con_Rel::is_included()
                : Poly_Con_Rel::is_disjoint();
            } else {
              if ((sp_sign > 0 && res == Poly_Con_Rel::is_disjoint())
                  ||
                  (sp_sign < 0 && res.implies(Poly_Con_Rel::is_included())))
                return Poly_Con_Rel::strictly_intersects();
              if (sp_sign > 0)
                res = Poly_Con_Rel::is_included();
            }
            break;
          case Gen::POINT:
          case Gen::CLOSURE_POINT:
            // A non-saturating closure point is treated as a point.
            if (no_point_or_nonsat_ray) {
              no_point_or_nonsat_ray = false;
              if (sp_sign > 0)
                res = Poly_Con_Rel::is_included();
              else if (sp_sign < 0)
                res = Poly_Con_Rel::is_disjoint();
            } else {
              if ((sp_sign > 0 && res == Poly_Con_Rel::is_disjoint())
                  ||
                  (sp_sign < 0 && res.implies(Poly_Con_Rel::is_included())))
                return Poly_Con_Rel::strictly_intersects();
              if (sp_sign > 0)
                res = Poly_Con_Rel::is_included();
            }
            break;
          } // switch (g.type())
        }
      }
    }
    break;

  case Con::STRICT_INEQUALITY:
    {
      res = res && Poly_Con_Rel::is_disjoint();
      bool no_point_or_nonsat_ray = true;
      for (const auto& g : gs) {
        const int sp_sign = Scalar_Prod::sign(c, g);
        if (sp_sign == 0) {
          if (g.is_point()) {
            if (no_point_or_nonsat_ray)
              no_point_or_nonsat_ray = false;
            else if (res == Poly_Con_Rel::is_included())
              return Poly_Con_Rel::strictly_intersects();
          }
        } else {
          assert(sp_sign != 0);
          switch (g.type()) {
          case Gen::LINE:
            return Poly_Con_Rel::strictly_intersects();
          case Gen::RAY:
            if (no_point_or_nonsat_ray) {
              no_point_or_nonsat_ray = false;
              res = (sp_sign > 0)
                ? Poly_Con_Rel::is_included()
                : Poly_Con_Rel::is_disjoint();
            } else {
              if ((sp_sign > 0 && res.implies(Poly_Con_Rel::is_disjoint()))
                  ||
                  (sp_sign <= 0 && res == Poly_Con_Rel::is_included()))
                return Poly_Con_Rel::strictly_intersects();
              if (sp_sign < 0)
                res = Poly_Con_Rel::is_disjoint();
            }
            break;
          case Gen::POINT:
          case Gen::CLOSURE_POINT:
            if (no_point_or_nonsat_ray) {
              no_point_or_nonsat_ray = false;
              if (sp_sign > 0)
                res = Poly_Con_Rel::is_included();
              else if (sp_sign < 0)
                res = Poly_Con_Rel::is_disjoint();
            } else {
              if ((sp_sign > 0 && res.implies(Poly_Con_Rel::is_disjoint()))
                  ||
                  (sp_sign <= 0 && res == Poly_Con_Rel::is_included()))
                return Poly_Con_Rel::strictly_intersects();
              if (sp_sign < 0)
                res = Poly_Con_Rel::is_disjoint();
            }
            break;
          } // switch (g.type())
        }
      }
    }
    break;
  } // switch (c.type())
  return res;
}

} // namespace detail

Poly_Con_Rel
Poly_Impl::relation_with(const Con& c) const {
  assert(space_dim() >= c.space_dim());

  if (marked_empty())
    return Poly_Con_Rel::saturates()
      && Poly_Con_Rel::is_included()
      && Poly_Con_Rel::is_disjoint();

  if (space_dim() == 0) {
    if (c.is_inconsistent())
      return (c.is_strict_inequality() && c.inhomo_term() == 0)
        ? Poly_Con_Rel::saturates() && Poly_Con_Rel::is_disjoint()
        : Poly_Con_Rel::is_disjoint();
    else
      return (c.is_equality() || c.inhomo_term() == 0)
        ? Poly_Con_Rel::saturates() && Poly_Con_Rel::is_included()
        : Poly_Con_Rel::is_included();
  }

  ensure_valid_gens();
  if (marked_empty())
    return Poly_Con_Rel::saturates()
      && Poly_Con_Rel::is_included()
      && Poly_Con_Rel::is_disjoint();

  return detail::relation_with(gens(), c);
}


namespace detail {

template <typename Src_Row, typename Dst_Rows>
bool
compute_sat_row(const Src_Row& src, const Dst_Rows& dst_rows,
                Bit_Row& sat_row) {
  for (dim_type i = dst_rows.size(); i-- > 0; ) {
    int s = Scalar_Prod::sign(src, dst_rows[i]);
    if (s > 0)
      sat_row.set(i);
    else if (s < 0)
      return false;
    else if (is_strict_ineq_or_point(src.type())
             && is_strict_ineq_or_point(dst_rows[i].type()))
      return false;
  }
  return true;
}

inline bool
violates(const Gen& g, const Con& c) {
  auto s = Scalar_Prod::sign(g, c);
  if (g.is_line())
    return s != 0;
  switch (c.type()) {
  case Con::EQUALITY:
    return s != 0;
  case Con::NONSTRICT_INEQUALITY:
    return s < 0;
  case Con::STRICT_INEQUALITY:
  default:
    return g.is_point() ? (s <= 0) : (s < 0);
  }
}

inline bool
satisfied_by_all(const Con& c, const Gens& gs) {
  return std::none_of(gs.begin(), gs.end(),
                      [&c](const Gen& g) { return violates(g, c); });
}

inline bool
satisfied_by_all(const Topol topol, const Con& c,
                 const Poly_Impl::Sys<Gens>& gs) {
  const bool dont_check_ns
    = (topol == Topol::CLOSED)
    || !c.is_strict_inequality()
    || gs.ns_rows.empty();

  if (!satisfied_by_all(c, gs.sing_rows))
    return false;
  if (dont_check_ns)
    return satisfied_by_all(c, gs.sk_rows);

  // Here we have an NNC and a strict inequality;
  // we need to check gs.ns_rows too.
  Bit_Row sat_row;
  if (!compute_sat_row(c, gs.sk_rows, sat_row))
    return false;
  for (const auto& g_ns : gs.ns_rows) {
    if (!g_ns.impl().intersects(sat_row))
      return false;
  }
  // All generators satisfy c.
  return true;
}

inline bool
satisfies_all(const Gen& g, const Cons& cs) {
  return std::none_of(cs.begin(), cs.end(),
                      [&g](const Con& c) { return violates(g, c); });
}

inline bool
satisfies_all(const Topol topol, const Gen& g,
              const Poly_Impl::Sys<Cons>& cs) {
  const bool dont_check_ns
    = (topol == Topol::CLOSED) || !g.is_point() || cs.ns_rows.empty();

  if (!satisfies_all(g, cs.sing_rows))
    return false;
  if (dont_check_ns)
    return satisfies_all(g, cs.sk_rows);

  // Here we have an NNC and a point; we need to check cs.ns_rows too.
  Bit_Row sat_row;
  if (!compute_sat_row(g, cs.sk_rows, sat_row))
    return false;
  for (const auto& c_ns : cs.ns_rows) {
    if (!c_ns.impl().intersects(sat_row))
      return false;
  }
  // All constraints are satisfied.
  return true;
}

inline bool
satisfies_all(const Topol topol, const Gen& g,
              const Poly_Impl::Sys<Cons>& cs1,
              const Poly_Impl::Sys<Cons>& cs2) {
  return satisfies_all(topol, g, cs1) && satisfies_all(topol, g, cs2);
}

inline bool
satisfy_all(const Gens& gs, const Cons& cs) {
  return std::all_of(gs.begin(), gs.end(),
                     [&cs](const Gen& g) { return satisfies_all(g, cs); });
}

inline bool
satisfy_all(const Gens& gs, const Cons& cs, Sat& sat) {
  for (auto i = gs.size(); i-- > 0; ) {
    if (!compute_sat_row(gs[i], cs, sat[i]))
      return false;
  }
  return true;
}

inline bool
satisfy_all(const Topol topol,
            const Poly_Impl::Sys<Gens>& gs,
            const Poly_Impl::Sys<Cons>& cs) {
  const bool dont_check_ns = (topol == Topol::CLOSED)
    || (gs.ns_rows.empty() && cs.ns_rows.empty());

  // Check gs singular rows (i.e., lines);
  // no need to check them against cs.ns_rows.
  if (!satisfy_all(gs.sing_rows, cs.sing_rows))
    return false;
  if (!satisfy_all(gs.sing_rows, cs.sk_rows))
    return false;

  // Check gs skel rows.
  if (!satisfy_all(gs.sk_rows, cs.sing_rows))
    return false;
  if (dont_check_ns)
    return satisfy_all(gs.sk_rows, cs.sk_rows);

  Sat sat(gs.sk_rows.size(), cs.sk_rows.size());
  if (!satisfy_all(gs.sk_rows, cs.sk_rows, sat))
    return false;

  // Check (strict ineqs in) cs.ns_rows ...
  for (const auto& c_ns : cs.ns_rows) {
    // ... against points in gs.sk_rows,
    for (auto i = gs.sk_rows.size(); i-- > 0; ) {
      if (gs.sk_rows[i].is_point() && !c_ns.impl().intersects(sat[i]))
        return false;
    }
    // ... and against points in gs.ns_rows,
    for (const auto& g_ns : gs.ns_rows) {
      if (!c_ns.impl().intersects(sat_all(g_ns, sat)))
        return false;
    }
  }
  // Check strict ineqs in cs.sk_rows against (points in) gs.ns_rows.
  for (auto j = cs.sk_rows.size(); j-- > 0; ) {
    if (cs.sk_rows[j].is_strict_inequality()) {
      Index_Set j_set;
      for (auto i = gs.sk_rows.size(); i-- > 0; ) {
        if (!sat[i][j])
          j_set.set(i);
      }
      if (std::any_of(gs.ns_rows.begin(), gs.ns_rows.end(),
                      [&j_set](const Index_Set& ns) {
                        return subset_or_equal(ns, j_set);
                      }))
        return false;
    }
  }
  // All checks passed.
  return true;
}

inline bool
satisfy_all(const Topol topol,
            const Poly_Impl::Sys<Gens>& gs1,
            const Poly_Impl::Sys<Gens>& gs2,
            const Poly_Impl::Sys<Cons>& cs1,
            const Poly_Impl::Sys<Cons>& cs2) {
  return satisfy_all(topol, gs1, cs1)
    && satisfy_all(topol, gs1, cs2)
    && satisfy_all(topol, gs2, cs1)
    && satisfy_all(topol, gs2, cs2);
}

inline bool
is_included_in(const Poly_Impl& x, const Poly_Impl& y) {
  assert(x.topol == y.topol);
  assert(x.dim == y.dim);
  assert(!x.marked_empty() && !y.marked_empty());
  assert(x.dim > 0);
  // We need `x' generators.
  x.ensure_valid_gens();
  if (x.marked_empty())
    return true;
  // We need `y' constraints.
  y.ensure_valid_cons();
  // All `x' gens have to satisfy all `y' cons.
  return satisfy_all(x.topol, x.gs, x.gs_pending, y.cs, y.cs_pending);
}

TV_Bool
quick_equals(const Poly_Impl& x, const Poly_Impl& y) {
  assert(x.topol == y.topol);
  assert(x.dim == y.dim);
  assert(!x.marked_empty() && !y.marked_empty());
  assert(x.dim > 0);

  if (!x.marked_min() || !y.marked_min())
    return TV_Bool::DONT_KNOW;

  if (x.num_equals() != y.num_equals()
      || x.num_lines() != y.num_lines()
      || num_rows(x.cs.sk_rows) != num_rows(y.cs.sk_rows)
      || num_rows(x.gs.sk_rows) != num_rows(y.gs.sk_rows)
      || num_rows(x.cs.ns_rows) != num_rows(y.cs.ns_rows)
      || num_rows(x.gs.ns_rows) != num_rows(y.gs.ns_rows))
    return TV_Bool::FALSE;
  /* FIXME.
     In order to enable the following code,
     we need to implement the sorting of the systems
     keeping ns_rows up-to-date.
  if (x.num_equals() == 0) {
    x.cs.sort();
    y.cs.sort();
    return (x.cs == y.cs) ? TV_Bool::TRUE : TV_Bool::FALSE;
  }
  if (x.num_lines() == 0) {
    x.gs.sort();
    y.gs.sort();
    return (x.gs == y.gs) ? TV_Bool::TRUE : TV_Bool::FALSE;
  }
  */
  return TV_Bool::DONT_KNOW;
}

} // namespace detail

Poly_Gen_Rel
Poly_Impl::relation_with(const Gen& g) const {
  assert(space_dim() >= g.space_dim());
  if (is_empty())
    return Poly_Gen_Rel::nothing();
  if (space_dim() == 0)
    return Poly_Gen_Rel::subsumes();
  ensure_valid_cons();
  assert(!marked_empty());
  return detail::satisfies_all(topol, g, cs, cs_pending)
    ? Poly_Gen_Rel::subsumes()
    : Poly_Gen_Rel::nothing();
}

void
Poly_Impl::add_con(Con c) {
  assert(c.space_dim() <= space_dim());
  if (marked_empty())
    return;
  if (space_dim() == 0
      || (c.is_strict_inequality() && is_necessarily_closed())) {
    assert(c.is_tautological() || c.is_inconsistent());
    if (c.is_inconsistent())
      set_empty();
    return;
  }
  ensure_valid_cons();
  assert(marked_min() || has_cs_pending());
  auto& dest = c.is_equality() ? cs_pending.sing_rows : cs_pending.sk_rows;
  dest.push_back(std::move(c));
  if (marked_min())
    set_status(Status::PENDING);
  assert(check_inv());
}

void
Poly_Impl::add_gen(Gen g) {
  assert(g.space_dim() <= space_dim());
  assert(!(is_necessarily_closed() && g.is_closure_point()));
  assert(!(space_dim() == 0 && g.is_line_or_ray()));

  if (space_dim() == 0) {
    if (marked_empty()) {
      assert(g.is_point());
      set_universe();
    }
    assert(check_inv());
    return;
  }
  ensure_valid_gens();

  if (marked_empty()) {
    assert(g.is_point());
    Sys<Gens> tmp;
    tmp.sk_rows.push_back(std::move(g));
    reinit_with_gens(dim, topol, std::move(tmp));
    return;
  }
  assert(marked_min() || has_gs_pending());
  auto& dest = g.is_line() ?  gs_pending.sing_rows : gs_pending.sk_rows;
  dest.push_back(std::move(g));
  if (marked_min())
    set_status(Status::PENDING);
  assert(check_inv());
}

void
Poly_Impl::minimize() const {
  if (marked_empty() || marked_min() || (space_dim() == 0))
    return;

  // The polyhedron is only `logically' const.
  auto& x = const_cast<Poly_Impl&>(*this);

  if (has_valid_cons()) {
    if (sat_c.num_rows() == 0)
      x.sat_c.transpose_assign(x.sat_g);
    if (detail::add_and_minimize<true>(x.dim, x.cs, x.cs_pending,
                                       x.gs, x.sat_c, x.sat_g))
      x.set_empty();
    else
      x.set_status(Status::MINIMIZED);
    assert(check_inv(false));
    return;
  }

  assert(has_valid_gens());
  if (sat_g.num_rows() == 0)
    x.sat_g.transpose_assign(x.sat_c);
  if (detail::add_and_minimize<false>(x.dim, x.gs, x.gs_pending,
                                      x.cs, x.sat_g, x.sat_c))
    assert(false);
  else
    x.set_status(Status::MINIMIZED);
  assert(check_inv(false));
}

namespace detail {

// If the transformation is not invertible we may transform
// valid lines and rays into the origin of the space.
// Fixing up the system is up to the caller.
void
affine_image_unsafe(Gens& gs, Var var, const Linear_Expr& expr,
                    const Integer& inhomo, const Integer& den) {
  assert(den > 0);
  Integer num;
  for (auto& g : gs) {
    Scalar_Prod::assign(num, expr, inhomo, g);
    auto& g_expr = g.impl().expr;
    auto& g_inhomo = g.impl().inhomo;
    if (den != 1) {
      g_expr *= den;
      g_inhomo *= den;
    }
    g_expr.set(var, num);
    g.strong_normalize();
  }
}

void
affine_image(Poly_Impl::Sys<Gens>& gs, Var var,
             Linear_Expr expr, Integer inhomo, Integer den) {
  if (den < 0) {
    neg_assign(expr);
    neg_assign(inhomo);
    neg_assign(den);
  }
  affine_image_unsafe(gs.sing_rows, var, expr, inhomo, den);
  affine_image_unsafe(gs.sk_rows, var, expr, inhomo, den);
  const bool invertible = (expr.get(var) != 0);
  if (!invertible) {
    Index_Set tbr = invalid_lines_and_rays(gs.sing_rows);
    erase_using_sorted_indices(gs.sing_rows, tbr);
    tbr = invalid_lines_and_rays(gs.sk_rows);
    if (!tbr.empty()) {
      erase_using_sorted_indices(gs.sk_rows, tbr);
      for (auto& ns : gs.ns_rows)
        ns.remove_all(tbr.begin(), tbr.end());
    }
  }
}

void
affine_preimage(Cons& cs, Var var, const Linear_Expr& expr,
                const Integer& inhomo, const Integer& den) {
  assert(den > 0);
  const auto& var_e = expr.get(var);
  for (auto& c : cs) {
    if (c.coeff(var) == 0)
      continue;
    auto var_c = c.coeff(var);
    auto& c_expr = c.impl().expr;
    auto& c_inhomo = c.impl().inhomo;
    if (den != 1) {
      c_expr *= den;
      c_inhomo *= den;
    }
    // FIXME: CHECKME.
    add_mul_assign(c_expr, var_c, expr);
    add_mul_assign(c_inhomo, var_c, inhomo);
    c_expr.set(var, var_c * var_e);
    c.strong_normalize();
  }
}

void
affine_preimage(Poly_Impl::Sys<Cons>& cs, Var var,
                Linear_Expr expr, Integer inhomo, Integer den) {
  if (den < 0) {
    neg_assign(expr);
    neg_assign(inhomo);
    neg_assign(den);
  }
  affine_preimage(cs.sing_rows, var, expr, inhomo, den);
  affine_preimage(cs.sk_rows, var, expr, inhomo, den);
}

} // namespace detail

void
Poly_Impl::affine_image(const Var var, const Linear_Expr& expr,
                        const Integer& inhomo, const Integer& den) {
  assert(space_dim() > 0);
  assert(var.space_dim() <= space_dim());
  assert(expr.space_dim() <= space_dim());
  assert(den != 0);
  if (marked_empty())
    return;
  const bool invertible = (expr.get(var) != 0);
  if (invertible) {
    detail::affine_image(gs, var, expr, inhomo, den);
    if (!gs_pending.empty())
      detail::affine_image(gs_pending, var, expr, inhomo, den);
    // Build the inverse transformation.
    auto inv_den = expr.get(var);
    auto inv_expr = -expr;
    inv_expr.set(var, den);
    auto inv_inhomo = -inhomo;
    detail::affine_preimage(cs, var, inv_expr, inv_inhomo, inv_den);
    if (!cs_pending.empty())
      detail::affine_preimage(cs_pending, var, inv_expr, inv_inhomo, inv_den);
    assert(check_inv());
    return;
  }
  // The transformation is not invertible.
  assert(!invertible);
  unconstrain(var);
  add_con(den * var - expr == inhomo);
}

void
Poly_Impl::affine_preimage(const Var var, const Linear_Expr& expr,
                           const Integer& inhomo, const Integer& den) {
  assert(space_dim() > 0);
  assert(var.space_dim() <= space_dim());
  assert(expr.space_dim() <= space_dim());
  assert(den != 0);
  if (marked_empty())
    return;
  const bool invertible = (expr.get(var) != 0);
  if (invertible) {
    // Invertible: compute inverse and delegate to affine_image.
    auto inv_den = expr.get(var);
    auto inv_expr = -expr;
    inv_expr.set(var, den);
    auto inv_inhomo = -inhomo;
    affine_image(var, inv_expr, inv_inhomo, inv_den);
    return;
  }
  // The transformation is not invertible.
  assert(!invertible);
  add_con(den * var - expr == inhomo);
  unconstrain(var);
}

namespace detail {

bool
check_par_affine(dim_type space_dim,
                 const std::vector<Var>& vars,
                 const std::vector<Linear_Expr>& exprs,
                 const Integers& inhomos, const Integers& dens) {
  const auto n_vars = vars.size();
  if (n_vars == 0)
    return false;
  if (n_vars != exprs.size()
      || n_vars != inhomos.size() || n_vars != dens.size())
    return false;

  std::vector<bool> is_target(space_dim, false);
  for (auto j = n_vars; j-- > 0; ) {
    if (vars[j].space_dim() > space_dim)
      return false;
    if (exprs[j].space_dim() > space_dim)
      return false;
    if (dens[j] == 0)
      return false;
    const auto v_id = vars[j].id();
    if (is_target[v_id])
      return false;
    is_target[v_id] = true;
  }
  return true;
}

bool
is_confluent(const dim_type space_dim,
             const std::vector<Var>& vars,
             const std::vector<Linear_Expr>& exprs) {
  const auto n_vars = num_rows(vars);
  if (n_vars < 2)
    return true;
  std::vector<bool> is_target(space_dim, false);
  for (auto j = n_vars; j-- > 0; ) {
    const auto v_id = vars[j].id();
    assert(v_id < space_dim && !is_target[v_id]);
    is_target[v_id] = true;
  }
  for (auto j = n_vars; j-- > 0; ) {
    const auto v_id = vars[j].id();
    for (auto i = exprs[j].space_dim(); i-- > 0; ) {
      if (is_target[i] && i != v_id)
        return false;
    }
  }
  return true;
}

void
par_affine_image_unsafe(Gens& gens,
                        const std::vector<Var>& vars,
                        const std::vector<Linear_Expr>& exprs,
                        const Integers& inhomos,
                        const Integer& den_lcm,
                        const Integers& mults,
                        const bool is_fractional,
                        const bool is_confluent) {
  // Note: preconditions have been checked by the caller.
  const auto n_vars = num_rows(vars);
  const auto& src = gens;
  Gens tmp = is_confluent ? Gens() : src;
  Gens& dst = is_confluent ? gens : tmp;

  const auto n_rows = num_rows(src);
  // Compute the numerators of the affine transformations
  // and assign them to the columns of `dst' indexed by `vars'.
  Integers nums(n_vars);
  for (auto i = n_rows; i-- > 0; ) {
    const auto& src_g = src[i];
    for (auto j = n_vars; j-- > 0; )
      Scalar_Prod::assign(nums[j], exprs[j], inhomos[j], src_g);
    auto& dst_expr = dst[i].linear_expr();
    auto& dst_inhomo = dst[i].impl().inhomo;
    if (is_fractional) {
      // Fractional: adjust numerators.
      dst_expr *= den_lcm;
      dst_inhomo *= den_lcm;
      for (auto j = n_vars; j-- > 0; ) {
        nums[j] *= mults[j];
        dst_expr.set(vars[j], nums[j]);
      }
    } else {
      // Not fractional: simply assign the numerators.
      for (auto j = n_vars; j-- > 0; )
        dst_expr.set(vars[j], nums[j]);
    }
    dst[i].strong_normalize();
  }
  if (!is_confluent)
    gens = std::move(dst);
}

void
par_affine_image(Poly_Impl::Sys<Gens>& gs,
                 const std::vector<Var>& vars,
                 const std::vector<Linear_Expr>& exprs,
                 const Integers& inhomos,
                 const Integer& den_lcm,
                 const Integers& mults,
                 const bool is_fractional,
                 const bool is_confluent,
                 const bool is_invertible) {
  par_affine_image_unsafe(gs.sing_rows, vars, exprs, inhomos,
                          den_lcm, mults, is_fractional, is_confluent);
  par_affine_image_unsafe(gs.sk_rows, vars, exprs, inhomos,
                          den_lcm, mults, is_fractional, is_confluent);
  if (!is_invertible) {
    Index_Set tbr = invalid_lines_and_rays(gs.sing_rows);
    erase_using_sorted_indices(gs.sing_rows, tbr);
    tbr = invalid_lines_and_rays(gs.sk_rows);
    if (!tbr.empty()) {
      erase_using_sorted_indices(gs.sk_rows, tbr);
      for (auto& ns : gs.ns_rows)
        ns.remove_all(tbr.begin(), tbr.end());
    }
  }
}

} // namespace detail

void
Poly_Impl::parallel_affine_image(const std::vector<Var>& vars,
                                 std::vector<Linear_Expr>& exprs,
                                 Integers& inhomos, Integers& dens) {
  assert(detail::check_par_affine(dim, vars, exprs, inhomos, dens));
  const auto n_vars = num_rows(vars);
  if (n_vars == 1) {
    affine_image(vars[0], exprs[0], inhomos[0], dens[0]);
    return;
  }

  ensure_valid_gens();
  if (marked_empty())
    return;
  // Ensure all dens are positive.
  for (auto j = n_vars; j-- > 0; ) {
    if (dens[j] < 0) {
      neg_assign(exprs[j]);
      neg_assign(inhomos[j]);
      neg_assign(dens[j]);
    }
  }
  // Compute lcm of denominators.
  Integer den_lcm = dens[n_vars - 1];
  for (auto j = n_vars - 1; j-- > 0; )
    lcm_assign(den_lcm, den_lcm, dens[j]);
  bool is_fractional = (den_lcm != 1);
  // Compute multipliers.
  Integers mults;
  if (is_fractional) {
    mults.resize(n_vars);
    for (auto j = n_vars; j-- > 0; )
      exact_div_assign(mults[j], den_lcm, dens[j]);
  }
  bool is_confluent = detail::is_confluent(dim, vars, exprs);
  bool is_invertible = true;
  for (auto j = n_vars; j-- > 0; ) {
    if (exprs[j].get(vars[j]) == 0) {
      is_invertible = false;
      break;
    }
  }
  Sys<Gens> tmp = std::move(gs);
  detail::concat_sys(tmp, std::move(gs_pending));
  detail::par_affine_image(tmp, vars, exprs, inhomos, den_lcm, mults,
                           is_fractional, is_confluent, is_invertible);
  reinit_with_gens(dim, topol, std::move(tmp));
  assert(check_inv());
}

namespace detail {

// Remove strict inequalities preserving the ray-face-cutter.
void
safe_remove_strict_ineqs(Cons& cs_sk, NS_Rows& cs_ns,
                         const Gens& gs_sk, const Sat& sat_c) {
  // Remove skeleton strict ineqs, except for positivity, if present.
  strict_ineqs_become_nonstrict_ineqs(cs_sk);
  // Remove non-skeleton strict ineqs
  cs_ns.clear();

  auto pos_i = std::find_if(cs_sk.begin(), cs_sk.end(),
                            std::mem_fn(&Con::is_tautological));
  const bool has_pos = (pos_i != cs_sk.end());
  if (has_pos) {
    // Positivity stays strict.
    assert(has_strict_ineq(cs_sk));
    return;
  }

  // Here positivity is not present in cs_sk.
  // We must reintroduce the rfc in cs_ns.
  Index_Set rays;
  for (auto g = num_rows(gs_sk); g-- > 0; )
    if (gs_sk[g].is_ray())
      rays.set(g);
  Bit_Row rays_sat = detail::sat_all(rays, sat_c);
  rays_sat.complement_until(num_rows(cs_sk));
  Index_Set efc(std::move(rays_sat));
  cs_ns.push_back(std::move(efc));
}

} // namespace detail

void
Poly_Impl::topological_closure_assign() {
  if (is_necessarily_closed() || marked_empty() || (space_dim() == 0))
    return;
  ensure_valid_gens();
  if (marked_empty())
    return;
  gs.ns_rows.clear();
  detail::closure_points_become_points(gs.sk_rows);
  gs_pending.ns_rows.clear();
  detail::closure_points_become_points(gs_pending.sk_rows);
  assert(!has_cs_pending());
  if (marked_min() || has_gs_pending()) {
    detail::safe_remove_strict_ineqs(cs.sk_rows, cs.ns_rows,
                                     gs.sk_rows, sat_c);
  }
  assert(check_inv());
}

void
Poly_Impl::intersection_assign(const Poly_Impl& y) {
  auto& x = *this;
  assert(x.topol == y.topol);
  assert(x.dim == y.dim);
  if (x.marked_empty())
    return;
  if (y.marked_empty()) {
    x.set_empty();
    return;
  }
  // Handle 0-dim universe case.
  if (x.dim == 0)
    return;

  x.ensure_valid_cons();
  assert(!x.marked_empty());
  y.minimize();
  assert(y.check_inv());
  if (y.marked_empty()) {
    x.set_empty();
    return;
  }
  assert(x.marked_min() || x.has_cs_pending());
  assert(y.marked_min());
  detail::concat_sys(x.cs_pending, y.cs);

  x.set_status(Status::PENDING);
  assert(x.check_inv());
}

void
Poly_Impl::poly_hull_assign(const Poly_Impl& y) {
  auto& x = *this;
  assert(x.topol == y.topol);
  assert(x.dim == y.dim);
  if (x.marked_empty()) {
    x = y;
    return;
  }
  if (x.dim == 0 || y.marked_empty())
    return;

  x.ensure_valid_gens();
  if (x.marked_empty()) {
    x = y;
    return;
  }
  y.minimize();
  assert(y.check_inv());
  if (y.marked_empty())
    return;
  assert(x.marked_min() || x.has_gs_pending());
  assert(y.marked_min());
  detail::concat_sys(x.gs_pending, y.gs);

  x.set_status(Status::PENDING);
  assert(x.check_inv());
}

void
Poly_Impl::poly_difference_assign(const Poly_Impl& y) {
  auto& x = *this;

  if (x.marked_empty() || y.marked_empty())
    return;
  if (x.space_dim() == 0 || y.contains(x)) {
    x.set_empty();
    return;
  }

  y.minimize();
  if (y.is_empty())
    return;
  x.minimize();

  const bool is_nnc = !x.is_necessarily_closed();
  Poly_Impl result(x.space_dim(), Spec_Elem::EMPTY, x.topol);
  for (const auto& c : y.cons()) {
    assert(!c.is_tautological() && !c.is_inconsistent());
    if (x.relation_with(c).implies(Poly_Con_Rel::is_included()))
      continue;
    // Build the complement of constraint c.
    auto expr = c.linear_expr();
    neg_assign(expr);
    auto inhomo = c.inhomo_term();
    neg_assign(inhomo);
    Con::Type type;
    switch (c.type()) {
    case Con::NONSTRICT_INEQUALITY:
      type = is_nnc ? Con::STRICT_INEQUALITY : Con::NONSTRICT_INEQUALITY;
      break;
    case Con::STRICT_INEQUALITY:
      type = Con::NONSTRICT_INEQUALITY;
      break;
    case Con::EQUALITY:
    default:
      if (is_nnc) {
        type = Con::STRICT_INEQUALITY;
        auto w = x;
        w.add_con(Con(c.linear_expr(), c.inhomo_term(), type));
        result.poly_hull_assign(w);
      } else {
        // We have already filtered out the case
        // when `x' is included in `y': the result is `x'.
        return;
      }
      break;
    }
    auto z = x;
    z.add_con(Con(std::move(expr), std::move(inhomo), type));
    result.poly_hull_assign(z);
  }
  x = std::move(result);
  assert(x.check_inv());
}

void
Poly_Impl::concatenate_assign(const Poly_Impl& y) {
  auto& x = *this;
  assert(x.topol == y.topol);
  if (x.marked_empty() || y.marked_empty()) {
    x.dim += y.dim;
    x.set_empty();
    return;
  }
  if (y.space_dim() == 0)
    return;
  if (x.space_dim() == 0) {
    x = y;
    return;
  }

  Cons y_cs = y.copy_cons();
  for (auto& y_c : y_cs)
    y_c.shift_space_dims(Var(0), x.space_dim());

  x.add_space_dims(y.space_dim());
  x.add_cons(std::move(y_cs));
  assert(x.check_inv());
}

void
Poly_Impl::time_elapse_assign(const Poly_Impl& y) {
  Poly_Impl& x = *this;
  assert(x.topol == y.topol && x.dim == y.dim);
  // Deal with the zero-dim special case.
  if (x.dim == 0) {
    if (y.marked_empty())
      x.set_empty();
    return;
  }
  // Deal with the empty special case.
  if (x.marked_empty() || y.marked_empty()
      || x.is_empty() || y.is_empty()) {
    x.set_empty();
    return;
  }

  assert(x.has_valid_gens() && y.has_valid_gens());
  auto& x_lines = x.gs_pending.sing_rows;
  std::copy(y.gs.sing_rows.begin(), y.gs.sing_rows.end(),
            std::back_inserter(x_lines));
  std::copy(y.gs_pending.sing_rows.begin(), y.gs_pending.sing_rows.end(),
            std::back_inserter(x_lines));
  auto& x_sk = x.gs_pending.sk_rows;
  detail::add_as_rays(y.gs.sk_rows, x_sk);
  detail::add_as_rays(y.gs_pending.sk_rows, x_sk);
  // Note: y non-skel points can be ignored.
  if (x.marked_min() && !x.gs_pending.empty())
    x.set_status(Status::PENDING);
  assert(x.check_inv());
}

void
Poly_Impl::widening_assign(const Poly_Impl& y, widen_type widen,
                           const Cons* upto_ptr) {
  auto& x = *this;
  assert(x.topol == y.topol);
  assert(x.dim == y.dim);
  // Filter out empty and zero dim cases.
  if (x.dim == 0 || x.marked_empty() || y.marked_empty())
    return;
  y.minimize();
  if (y.marked_empty())
    return;
  x.minimize();
  assert(!x.marked_empty());

  Index_Set upto;
  if (upto_ptr) {
    // Before modifying x, select from upto_cs those constraints
    // that are satisfied by x.
    const auto& upto_cs = *upto_ptr;
    for (dim_type i = upto_cs.size(); i-- > 0; ) {
      if (detail::satisfied_by_all(x.topol, upto_cs[i], x.gs))
        upto.set(i);
    }
  }

  // Apply specific widening algorithm.
  widen(x, y);

  if (upto_ptr) {
    // Now add selected upto constraints.
    const auto& upto_cs = *upto_ptr;
    for (auto i : upto)
      x.add_con(upto_cs[i]);
  }
}

void
Poly_Impl::add_space_dims(dim_type m, bool project) {
  if (m == 0)
    return;
  if (marked_empty()) {
    dim += m;
    return;
  }
  if (project) {
    /* Projecting on zero. */
    ensure_valid_cons();
    auto& eqs = cs.sing_rows;
    eqs.reserve(eqs.size() + m);
    for (dim_type i = 0; i < m; ++i)
      eqs.push_back(Var(dim + i) == 0);
    dim += m;
    assert(check_inv());
    return;
  }
  /* Embedding. */
  ensure_valid_gens();
  if (marked_empty()) {
    dim += m;
    return;
  }
  auto& lines = gs.sing_rows;
  lines.reserve(dim + m);
  for (dim_type i = 0; i < m; ++i)
    lines.push_back(line(Var(dim + i)));
  dim += m;
  assert(check_inv());
}

void
Poly_Impl::remove_higher_space_dims(dim_type new_dim) {
  assert(new_dim <= space_dim());
  if (space_dim() == new_dim)
    return;
  ensure_valid_gens();
  if (marked_empty()) {
    dim = new_dim;
    assert(check_inv());
    return;
  }
  assert(has_valid_gens());
  if (new_dim == 0) {
    dim = new_dim;
    set_universe();
    return;
  }

  // Move all generators into tmp gens.
  Sys<Gens> tmp = std::move(gs);
  detail::concat_sys(tmp, std::move(gs_pending));

  Index_Set tbr;
  // Process lines.
  erase_higher_dims(tmp.sing_rows, new_dim);
  tbr = invalid_lines_and_rays(tmp.sing_rows);
  erase_using_sorted_indices(tmp.sing_rows, tbr);
  // Process skel generators.
  erase_higher_dims(tmp.sk_rows, new_dim);
  tbr = invalid_lines_and_rays(tmp.sk_rows);
  if (!tbr.empty()) {
    erase_using_sorted_indices(tmp.sk_rows, tbr);
    // Process non-skel generators.
    for (auto& ns : tmp.ns_rows)
      ns.remove_all(tbr.begin(), tbr.end());
  }
  reinit_with_gens(new_dim, topol, std::move(tmp));
}

namespace detail {

Linear_Expr
map_space_dims(Linear_Expr&& src, const Dims& pfunc, dim_type new_dim) {
  Linear_Expr res;
  res.set_space_dim(new_dim);
  const auto i_end = std::min(src.space_dim(), num_rows(pfunc));
  for (dim_type i = 0; i != i_end; ++i) {
    if (pfunc[i] != not_a_dim())
      res[pfunc[i]] = std::move(src[i]);
  }
  return res;
}

void
map_space_dims_unsafe(Gens& gens, const Dims& pfunc, dim_type new_dim) {
  // Note: this function may create invalid rays/lines.
  // We do not eliminate them here (it is done by the caller),
  // because we need to keep the non-skel component consistent.
  for (auto& g : gens) {
    auto& e = g.linear_expr();
    e = map_space_dims(std::move(e), pfunc, new_dim);
    g.strong_normalize();
  }
}

void
map_space_dims(Poly_Impl::Sys<Gens>& sys,
               const Dims& pfunc, dim_type new_dim) {
  Index_Set tbr;
  // Process lines.
  map_space_dims_unsafe(sys.sing_rows, pfunc, new_dim);
  tbr = invalid_lines_and_rays(sys.sing_rows);
  erase_using_sorted_indices(sys.sing_rows, tbr);
  // Process skel and non-skel generators.
  map_space_dims_unsafe(sys.sk_rows, pfunc, new_dim);
  tbr = invalid_lines_and_rays(sys.sk_rows);
  if (!tbr.empty()) {
    erase_using_sorted_indices(sys.sk_rows, tbr);
    for (auto& ns : sys.ns_rows)
      ns.remove_all(tbr.begin(), tbr.end());
  }
  promote_singletons(sys.ns_rows, sys.sk_rows);
}

} // namespace detail

void
Poly_Impl::map_space_dims(const Dims& pfunc) {
  assert(check_inv());
  assert(space_dim() == num_rows(pfunc));
  if (dim == 0)
    return;
  const auto max_dim = *std::max_element(pfunc.begin(), pfunc.end());
  if (max_dim == not_a_dim()) {
    const bool nonempty = !is_empty();
    dim = 0;
    if (nonempty)
      set_universe();
    assert(check_inv());
    return;
  }
  const auto new_dim = 1 + max_dim;
  if (new_dim == dim) {
    // pfunc is a permutation.
    if (marked_empty())
      return;
    Dims cycle;
    cycle.reserve(dim);
    std::vector<bool> visited(dim);
    for (dim_type i = dim; i-- > 0; ) {
      if (visited[i])
        continue;
      dim_type j = i;
      do {
        visited[j] = true;
        const dim_type k = pfunc[j];
        assert(k != not_a_dim());
        if (k == j)
          break;
        cycle.push_back(j);
        j = k;
      } while (!visited[j]);
      using detail::permute_space_dims;
      if (!cs.empty())
        permute_space_dims(cs, cycle);
      if (!cs_pending.empty())
        permute_space_dims(cs_pending, cycle);
      if (!gs.empty())
        permute_space_dims(gs, cycle);
      if (!gs_pending.empty())
        permute_space_dims(gs_pending, cycle);
      // sat matrices are unaffected.
      cycle.clear();
    }
    assert(check_inv());
    return;
  }

  // `pfunc' is not a permutation.
  if (is_empty()) {
    dim = new_dim;
    assert(check_inv());
    return;
  }
  assert(has_valid_gens());
  Sys<Gens> tmp = std::move(gs);
  detail::concat_sys(tmp, std::move(gs_pending));
  detail::map_space_dims(tmp, pfunc, new_dim);
  reinit_with_gens(new_dim, topol, std::move(tmp));
}

namespace detail {

template <typename Rows>
bool
check_inv(const Poly_Impl::Sys<Rows>& sys, dim_type dim) {
  using Row = typename Rows::value_type;
  if (std::any_of(sys.sing_rows.begin(), sys.sing_rows.end(),
                  [dim](const Row& sg) {
                    return !is_singular(sg)
                      || (sg.space_dim() > dim)
                      || !(sg.check_inv());
                  }))
    return false;
  if (std::any_of(sys.sk_rows.begin(), sys.sk_rows.end(),
                  [dim](const Row& sk) {
                    return is_singular(sk)
                      || (sk.space_dim() > dim)
                      || !(sk.check_inv());
                  }))
    return false;
  const auto sk_size = num_rows(sys.sk_rows);
  if (std::any_of(sys.ns_rows.begin(), sys.ns_rows.end(),
                  [sk_size](const Index_Set& ns) {
                    return (ns.size() < 2) || (ns.last() >= sk_size);
                  }))
    return false;
  // All checks passed.
  return true;
}

/**
 * There must always be a cutter.
 * - If there are rays, there must be the ray-face-cutter,
 * i.e., a constraint such that: sat_g(rfc) = R.
 * If present, the strict positivity is a valid rfc.
 * If implied by some constraints in cs_sk, and therefore not
 * present in cs_sk, these constraints are the rfc.
 * - If there aren't rays, the efc is the maximal one.
 */
bool
check_efc(const NS_Rows& cs_ns,
          const Cons& cs_sk,
          const Gens& gs_sk,
          const Sat& sat_g, const Sat& sat_c) {
  if (!detail::check_cutters(cs_ns, cs_sk, gs_sk, sat_g))
    return false;

  auto efc_i = std::find_if(cs_ns.cbegin(), cs_ns.cend(),
                            [&sat_g, &gs_sk](const Index_Set& ns) {
                              return is_empty_face_cutter(ns, sat_g, gs_sk);
                            });
  // We already know that there is a valid cutter.
  if (efc_i == cs_ns.cend())
    return true;

  // Check that the efc is the correct rfc.
  // Compute the expected rfc.
  Index_Set rays;
  for (auto g = num_rows(gs_sk); g-- > 0; )
    if (gs_sk[g].is_ray())
      rays.set(g);
  Bit_Row rfc_sat = sat_all(rays, sat_c);
  rfc_sat.complement_until(num_rows(cs_sk));
  // Note: if `rays' is empty, this computes the maximal set.
  Index_Set real_efc(std::move(rfc_sat));

  return (real_efc == *efc_i);
}

} // namespace detail

bool
Poly_Impl::check_inv(bool do_equals) const {
  std::string reason;
#ifdef NDEBUG
  auto maybe_dump = []() {};
#else // In debugging mode, be noisy.
  auto maybe_dump = [this, &reason]() {
    std::cerr << reason << std::endl;
    this->ascii_dump(std::cerr);
    std::cerr << reason << std::endl;
  };
#endif

  /* Check 0-dim polyhedra. */
  if (dim == 0 && !(marked_empty() || marked_min())) {
    reason = "Zero-dim poly neither empty nor minimized";
    maybe_dump();
    return false;
  }

  /* Check systems cardinalities. */
  if (!detail::check_inv(cs, dim))
    reason = "non-pending constraints are broken";
  else if (!detail::check_inv(gs, dim))
    reason = "non-pending generators are broken";
  else if (!detail::check_inv(cs_pending, dim))
    reason = "pending constraints are broken";
  else if (!detail::check_inv(gs_pending, dim))
    reason = "pending generators are broken";
  if (!reason.empty()) {
    maybe_dump();
    return false;
  }

  // Checks status consistency.
  switch (status) {

  case Status::EMPTY:
    if (not ((cs.sing_rows.size() == 1)
             && cs.sing_rows[0].space_dim() == 0
             && cs.sk_rows.empty() && cs.ns_rows.empty()
             && cs_pending.empty()
             && gs.empty() && gs_pending.empty()
             && sat_c.empty() && sat_g.empty()))
      reason = "poly marked empty is broken";
    break;

  case Status::PENDING:
    if (cs_pending.empty() && gs_pending.empty()) {
      reason = "poly marked pending has no pending cons/gens";
      break;
    } else if (!cs_pending.empty() && !gs_pending.empty()) {
      reason = "poly marked pending has both pending cons and gens";
      break;
    } else if (cs.empty()) {
      reason = "poly marked pending has empty non-pending cons";
      break;
    } else if (gs.empty()) {
      reason = "poly marked pending has empty non-pending gens";
      break;
    }
    // Intentionally falling through:
    // check minimization invariants for the non-pending part.

    [[ fallthrough ]];
  case Status::MINIMIZED:
    {
      if (not (num_rows(cs.sing_rows) <= dim
               && num_rows(gs.sing_rows) <= dim
               && !sat_c.empty() && !sat_g.empty()
               && (sat_c.num_rows() == sat_g.num_cols())
               && (sat_g.num_rows() == sat_c.num_cols())
               && (sat_c.num_rows() == num_rows(gs.sk_rows))
               && (sat_g.num_rows() == num_rows(cs.sk_rows)))) {
        reason = "minimized poly: broken cardinalities";
        break;
      }
      // Note: recheck status (may be falling through from PENDING).
      if (status == Status::MINIMIZED
          && !(cs_pending.empty() && gs_pending.empty())) {
        reason = "minimized poly has pending cons/gens";
        break;
      }
      // Check sat matrix consistency.
      Sat sat_c_tr;
      sat_c_tr.transpose_assign(sat_c);
      if (sat_g != sat_c_tr) {
        reason = "poly: sat_c is not the transpose of sat_g";
        break;
      }
      if (not detail::check_sat(sat_g, cs.sk_rows, gs.sk_rows)) {
        reason = "poly: sat matrix does not match scalar prod";
        break;
      }
      if (do_equals) {
        // Copy the non-pending part of this into x.
        Poly_Impl x = *this;
        x.cs_pending.clear();
        x.gs_pending.clear();
        x.status = Status::MINIMIZED;
        // Create y using the constraints of x as pending.
        Poly_Impl y(x.dim, Spec_Elem::UNIVERSE, x.topol);
        y.cs_pending = x.cs;
        y.set_status(Status::PENDING);
        y.minimize();
        if (!y.equals(x)) {
          reason = "poly said minimized, but it is not";
          break;
        }
      }
      if (!detail::check_efc(cs.ns_rows, cs.sk_rows,
                             gs.sk_rows, sat_g, sat_c)) {
        reason = "poly: invalid efc";
        break;
      }
    }
    break;

  default:
    reason = "poly: unknown status value";
    break;
  }

  if (!reason.empty())
    maybe_dump();

  return reason.empty();
}

void
Poly_Impl::ascii_dump(std::ostream& s) const {
  using namespace IO_Operators;
  s << "topol " << topol << "\n";
  s << "dim " << dim << "\n";
  s << "status " << status << "\n";
  s << "=> cs sys\n";
  cs.ascii_dump(s);
  s << "=> gs sys\n";
  gs.ascii_dump(s);
  s << "sat_c\n";
  sat_c.ascii_dump(s);
  s << "sat_g\n";
  sat_g.ascii_dump(s);
  s << "=> cs_pending\n";
  cs_pending.ascii_dump(s);
  s << "=> gs_pending\n";
  gs_pending.ascii_dump(s);
}

namespace detail {

template <typename Row>
bool
ascii_load(std::vector<Row>& rows, dim_type nrows, std::istream& s) {
  rows.resize(nrows);
  for (dim_type i = 0; i < nrows; ++i) {
    if (!rows[i].ascii_load(s))
      return false;
  }
  return true;
}

template <typename Rows>
bool
ascii_load(Poly_Impl::Sys<Rows>& sys, std::istream& s) {
  sys.clear();
  dim_type num = 0;
  if (not (ascii_load_string(s, "sing_rows")
           && (s >> num) && (num >= 0)
           && ascii_load(sys.sing_rows, num, s)))
    return false;
  if (not (ascii_load_string(s, "sk_rows")
           && (s >> num) && (num >= 0)
           && ascii_load(sys.sk_rows, num, s)))
    return false;
  if (not (ascii_load_string(s, "ns_rows")
           && (s >> num) && (num >= 0)
           && ascii_load(sys.ns_rows, num, s)))
    return false;
  return true;
}

} // namespace detail

bool
Poly_Impl::ascii_load(std::istream& s) {
  std::string str;
  if (not (ascii_load_string(s, "topol") && (s >> str)))
    return false;
  if (str == "CLOSED")
    topol = Topol::CLOSED;
  else if (str == "NNC")
    topol = Topol::NNC;
  else
    return false;

  if (not (ascii_load_string(s, "dim") && (s >> dim) && (dim >= 0)))
    return false;

  if (not (ascii_load_string(s, "status") && (s >> str)))
    return false;
  if (str == "EMPTY")
    status = Status::EMPTY;
  else if (str == "MINIMIZED")
    status = Status::MINIMIZED;
  else if (str == "PENDING")
    status = Status::PENDING;
  else
    return false;

  if (not (ascii_load_string(s, "=>")
           && ascii_load_string(s, "cs")
           && ascii_load_string(s, "sys")
           && detail::ascii_load(cs, s)))
    return false;
  if (not (ascii_load_string(s, "=>")
           && ascii_load_string(s, "gs")
           && ascii_load_string(s, "sys")
           && detail::ascii_load(gs, s)))
    return false;

  if (not (ascii_load_string(s, "sat_c") && sat_c.ascii_load(s)))
    return false;
  if (not (ascii_load_string(s, "sat_g") && sat_g.ascii_load(s)))
    return false;

  if (not (ascii_load_string(s, "=>")
           && ascii_load_string(s, "cs_pending")
           && detail::ascii_load(cs_pending, s)))
    return false;
  if (not (ascii_load_string(s, "=>")
           && ascii_load_string(s, "gs_pending")
           && detail::ascii_load(gs_pending, s)))
    return false;

  assert(check_inv());
  return true;
}

namespace IO_Operators {

std::ostream&
operator<<(std::ostream& s, Topol topol) {
  s << (topol == Topol::CLOSED ? "C" : "NNC");
  return s;
}

std::ostream&
operator<<(std::ostream& s, Poly_Impl::Status status) {
  using Status = Poly_Impl::Status;
  switch (status) {
  case Status::EMPTY:
    s << "EMPTY";
    break;
  case Status::MINIMIZED:
    s << "MINIMIZED";
    break;
  case Status::PENDING:
    s << "PENDING";
    break;
  default:
    assert(false);
    break;
  }
  return s;
}

std::ostream&
operator<<(std::ostream& s, const Poly_Impl& ph) {
  if (ph.is_empty()) {
    s << "false";
    return s;
  }
  ph.minimize();
  const auto& cs = ph.cons();
  auto i = cs.begin(), i_end = cs.end();
  bool comma = false;
  for (const auto& c : ph.cons()) {
    if (comma) s << ", ";
    s << c;
    comma = true;
  }
  return s;
}

std::ostream&
operator<<(std::ostream& s, const Poly& ph) {
  s << ph.impl();
  return s;
}

} // namespace IO_Operators

namespace detail {

// Just for debugging purposes
bool quick_equals_test(Poly_Impl x, Poly_Impl y) {
  x.minimize();
  y.minimize();
  // quick_equals assumes polyhedra are not 0-dim and are not empty.
  if (x.dim == 0 || x.is_empty() || y.is_empty())
    return true;
  auto res = quick_equals(x, y);
  if (res == TV_Bool::DONT_KNOW)
    return true;
  assert(res != TV_Bool::TRUE);
  // If quick_equals says false, it has to be correct.
  return (!is_included_in(y, x) || !is_included_in(x, y));
}

} // namespace detail

bool
Poly_Impl::equals(const Poly_Impl& y) const {
  const auto& x = *this;
  if (x.topol != y.topol || x.dim != y.dim)
    return false;

  if (x.marked_empty())
    return y.is_empty();
  else if (y.marked_empty())
    return x.is_empty();
  else if (x.space_dim() == 0)
    return true;

  assert(detail::quick_equals_test(x, y));
  switch (detail::quick_equals(x, y)) {
  case TV_Bool::TRUE:
    return true;
  case TV_Bool::FALSE:
    return false;
  case TV_Bool::DONT_KNOW:
  default:
    using detail::is_included_in;
    if (!is_included_in(x, y))
      return false;
    if (x.marked_empty())
      return y.is_empty();
    else
      return is_included_in(y, x);
  }
}

bool
Poly_Impl::contains(const Poly_Impl& y) const {
  const auto& x = *this;
  assert(x.topol == y.topol);
  assert(x.dim == y.dim);
  if (y.marked_empty())
    return true;
  if (x.marked_empty())
    return y.is_empty();
  if (y.space_dim() == 0)
    return true;
  if (detail::quick_equals(x, y) == TV_Bool::TRUE)
    return true;
  return detail::is_included_in(y, x);
}

bool
Poly_Impl::is_disjoint_from(const Poly_Impl& y) const {
  Poly_Impl z = *this;
  z.intersection_assign(y);
  return z.is_empty();
}

bool
Poly_Impl::bounds_expr(bool from_above, const Linear_Expr& expr) const {
  // A zero-dim polyhedron bounds everything.
  if (space_dim() == 0)
    return true;
  // An empty polyhedron bounds everything.
  if (is_empty())
    return true;

  // Predicate checking if a line/ray causes expr to be unbounded.
  auto makes_unbounded = [&expr, from_above](const Gen& g) {
    auto sp_sign = expr.scalar_prod_sign(g.linear_expr());
    if (sp_sign == 0)
      return false;
    if (g.is_line())
      return true;
    assert(g.is_ray());
    return static_cast<bool>(from_above ^ (sp_sign > 0));
  };

  // Check lines.
  if (std::any_of(gs.sing_rows.begin(),
                  gs.sing_rows.end(), makes_unbounded))
    return false;
  if (std::any_of(gs_pending.sing_rows.begin(),
                  gs_pending.sing_rows.end(), makes_unbounded))
    return false;
  // Check rays.
  for (const auto& g : gs.sk_rows) {
    if (g.is_ray() && makes_unbounded(g))
      return false;
  }
  for (const auto& g : gs_pending.sk_rows) {
    if (g.is_ray() && makes_unbounded(g))
      return false;
  }
  // expr is bounded.
  return true;
}

bool
Poly_Impl::constrains(Var var) const {
  assert(var.space_dim() <= space_dim());
  if (marked_empty())
    return true;

  auto c_pred = [var](const Con& c) {
    return c.linear_expr().get(var) != 0;
  };
  auto cs_pred = [c_pred](const Cons& cs) {
    return std::any_of(cs.begin(), cs.end(), c_pred);
  };
  if (marked_min())
    return cs_pred(cs.sing_rows) || cs_pred(cs.sk_rows);

  if (has_valid_gens()) {
    // Try a quick, incomplete check for the universe polyhedron.
    if (num_rows(gs.sing_rows) == space_dim())
      return false;
    // Predicate for lines.
    auto lines_pred = [var](const Gen& g) {
      assert(g.is_line());
      return g.linear_expr().all_zeroes_except(var);
    };
    // Scan lines.
    if (std::any_of(gs.sing_rows.begin(),
                    gs.sing_rows.end(), lines_pred))
      return false;
    // Scan pending lines.
    if (std::any_of(gs_pending.sing_rows.begin(),
                    gs_pending.sing_rows.end(), lines_pred))
      return false;

    // Scan skel (also pending) for opposite rays.
    bool has_pos_ray = false;
    bool has_neg_ray = false;
    // Predicate for rays (note: it has status).
    auto rays_pred = [var, &has_pos_ray, &has_neg_ray](const Gen& g) {
      if (g.is_ray() && g.linear_expr().all_zeroes_except(var)) {
        const int sign = sgn(g.coeff(var));
        if (sign > 0) {
          has_pos_ray = true;
          if (has_neg_ray)
            return true;
        } else {
          assert(sign < 0);
          has_neg_ray = true;
          if (has_pos_ray)
            return true;
        }
      }
      return false;
    };
    // Scan skel for opposite rays.
    if (std::any_of(gs.sk_rows.begin(), gs.sk_rows.end(), rays_pred))
      return false;
    // Scan pending skel for opposite rays.
    if (std::any_of(gs_pending.sk_rows.begin(),
                    gs_pending.sk_rows.end(), rays_pred))
      return false;
  }

  minimize();
  if (marked_empty())
    return true;
  return cs_pred(cs.sing_rows) || cs_pred(cs.sk_rows);
}

bool
Poly_Impl::min_max(const bool minimizing,
                   const Linear_Expr& expr, const Integer& inhomo,
                   Integer& num, Integer& den,
                   bool& included, Gen& g) const {
  assert(expr.space_dim() <= space_dim());
  if (space_dim() == 0) {
    if (marked_empty())
      return false;
    num = inhomo;
    den = 1;
    included = true;
    g = point();
    return true;
  }
  if (is_empty())
    return false;

  assert(has_valid_gens());

  // Check lines.
  auto check_line = [&expr](const Gen& g) {
    return expr.scalar_prod_sign(g.linear_expr()) != 0;
  };
  if (std::any_of(gs.sing_rows.begin(), gs.sing_rows.end(), check_line))
    return false;
  if (std::any_of(gs_pending.sing_rows.begin(), gs_pending.sing_rows.end(),
                  check_line))
    return false;

  // Check skel and non-skel generators.
  const Gen* cand = nullptr;
  Rational value;

  auto check_skel = [&](const Gen& g) {
    Integer sp;
    expr.scalar_prod_add(sp, g.linear_expr());
    if (g.is_ray()) {
      const int sp_sign = sgn(sp);
      if ((minimizing && sp_sign < 0) || (!minimizing && sp_sign > 0))
        return true;
    } else {
      assert(g.is_point_or_closure_point());
      Rational new_value(sp, g.divisor());
      const bool new_cand = (cand == nullptr)
        || (minimizing && new_value < value)
        || (!minimizing && new_value > value)
        || (g.is_point() && cand->is_closure_point() && new_value == value);
      if (new_cand) {
        cand = &g;
        value = new_value;
      }
    }
    return false;
  };
  // Skel component.
  if (std::any_of(gs.sk_rows.begin(), gs.sk_rows.end(), check_skel))
    return false;
  if (std::any_of(gs_pending.sk_rows.begin(), gs_pending.sk_rows.end(),
                  check_skel))
    return false;
  // Non-skel component: materialize.
  Gen mater_cand;
  for (const auto& ns : gs.ns_rows) {
    Gen mater = detail::materialize(ns, gs.sk_rows);
    if (check_skel(mater))
      assert(false);
    if (cand == &mater) {
      mater_cand = std::move(mater);
      cand = &mater_cand;
    }
  }
  for (const auto& ns : gs_pending.ns_rows) {
    Gen mater = detail::materialize(ns, gs_pending.sk_rows);
    if (check_skel(mater))
      assert(false);
    if (cand == &mater) {
      mater_cand = std::move(mater);
      cand = &mater_cand;
    }
  }

  assert(cand != nullptr);
  value += Rational(inhomo, 1);
  num = value.get_num();
  den = value.get_den();
  included = cand->is_point();
  g = *cand;
  return true;
}

namespace detail {

Index_Set
fold_gens(Gens& gens, const Index_Set& vars, Var v) {
  Gens new_gens;
  Index_Set invalid_gens;
  auto is_valid = [](const Gen& g) {
    return g.is_point_or_closure_point() || !g.linear_expr().is_zero();
  };
  const auto old_dst = v.id();
  const auto new_dst = old_dst
    - std::count_if(vars.begin(), vars.end(),
                    [old_dst](dim_type id) { return id < old_dst; });
  const auto min_dim = new_dst + 1;

  Index_Set vv = vars;
  vv.set(v.id());
  for (dim_type k = 0, k_end = num_rows(gens); k != k_end; ++k) {
    auto& g = gens[k];
    // Compute min and max coeffs for dimensions in vars.
    auto mm = std::minmax_element(vv.begin(), vv.end(),
                                  [&g](dim_type i, dim_type j) {
                                    return g.coeff(Var(i)) < g.coeff(Var(j));
                                  });
    // Note: copies are meant.
    Integer c_min = g.coeff(Var(*mm.first));
    Integer c_max = g.coeff(Var(*mm.second));
    g.linear_expr().remove_space_dims(vars.begin(), vars.end());
    if (g.space_dim() < min_dim)
      g.set_space_dim(min_dim);
    if (c_min != c_max) {
      // Note: copy has to be done *before* modifying g.
      Gen new_g = g;
      new_g.linear_expr()[new_dst] = std::move(c_max);
      if (is_valid(new_g)) {
        new_g.strong_normalize();
        new_gens.push_back(std::move(new_g));
      }
    }
    g.linear_expr()[new_dst] = std::move(c_min);
    if (is_valid(g))
      g.strong_normalize();
    else
      invalid_gens.set(k);
  }
  // Move new_gens into gens (at the end).
  gens.insert(gens.end(), new_gens.begin(), new_gens.end());
  return invalid_gens;
}

} // namespace detail

void
Poly_Impl::fold_space_dims(const Index_Set& vars, Var dst_var) {
  assert(dst_var.space_dim() <= space_dim());
  assert(!vars.test(dst_var.id()));
  if (vars.empty())
    return;
  assert(vars.last() < space_dim());

  // Generators are needed.
  ensure_valid_gens();
  // Update space dimension.
  dim -= vars.size();
  if (marked_empty())
    return;
  if (dim == 0) {
    set_universe();
    return;
  }

  // FIXME: speculative optimization.
  // If status == MINIMIZED, then we could check if the equalities
  // imply that all vars are equal to dst_var.
  // In this case, folding is a nop.

  // Move everything into new gs system.
  Sys<Gens> tmp = std::move(gs);
  detail::concat_sys(tmp, std::move(gs_pending));
  // Materialize non-skel generators.
  // CHECKME: is there any more efficient way?
  for (const auto& ns : tmp.ns_rows)
    tmp.sk_rows.push_back(detail::materialize(ns, tmp.sk_rows));
  tmp.ns_rows.clear();

  // Fold dims in geometric generators.
  Index_Set invalid = detail::fold_gens(tmp.sing_rows, vars, dst_var);
  erase_using_sorted_indices(tmp.sing_rows, invalid);
  invalid = detail::fold_gens(tmp.sk_rows, vars, dst_var);
  erase_using_sorted_indices(tmp.sk_rows, invalid);
  reinit_with_gens(dim, topol, std::move(tmp));
}

Poly_Impl
Poly_Impl::split(const Con& c, Topol t) {
  assert(t == Topol::CLOSED || t == topology());
  assert(!c.is_equality());
  assert(t == Topol::NNC || c.is_nonstrict_inequality());
  auto& ph_then = *this;
  auto ph_else = Poly_Impl(ph_then.space_dim(),
                           Spec_Elem::EMPTY,
                           ph_then.topology());
  ph_then.minimize();
  detail::split_poly<true>(ph_then, ph_else, c, t);
  assert(ph_then.is_minimized());
  assert(ph_else.is_minimized());
  return ph_else;
}

} // namespace pplite

