/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_Poly_hh
#define pplite_Poly_hh 1

#include "globals.hh"
#include "ascii_dump_load.hh"
#include "Var.hh"
#include "Linear_Expr.hh"
#include "Con.hh"
#include "Gen.hh"
#include "mater_iterator.hh"
#include "Index_Set.hh"
#include "Sat.hh"
#include "Poly_Rel.hh"

#include <tuple>
#include <vector>
#include <functional>
#include <iosfwd>
#include <cassert>

namespace pplite {

enum class Topol { CLOSED, NNC };
enum class Spec_Elem { EMPTY, UNIVERSE };

/* The implementation class: no encapsulation (everything public). */
struct Poly_Impl {
  enum class Status {
    EMPTY,
    MINIMIZED,
    PENDING
  };

  template <typename SK_Rows>
  struct Sys {
    using rows_type = SK_Rows;
    using row_type = typename SK_Rows::value_type;
    SK_Rows sing_rows;
    SK_Rows sk_rows;
    NS_Rows ns_rows;
    void clear() {
      sing_rows.clear();
      sk_rows.clear();
      ns_rows.clear();
    }
    bool empty() const {
      return sing_rows.empty() && sk_rows.empty();
    }
    void ascii_dump(std::ostream& s) const {
      s << "sing_rows " << sing_rows.size() << "\n";
      ascii_dump_all(s, sing_rows);
      s << "sk_rows " << sk_rows.size() << "\n";
      ascii_dump_all(s, sk_rows);
      s << "ns_rows " << ns_rows.size() << "\n";
      ascii_dump_all(s, ns_rows);
    }
    bool operator==(const Sys& y) const {
      return sing_rows == y.sing_rows
        && ns_rows == y.ns_rows
        && sk_rows == y.sk_rows;
    }
    void add(const row_type& new_elem) {
      if (detail::is_singular(new_elem))
        sing_rows.push_back(new_elem);
      else
        sk_rows.push_back(new_elem);
    }
  };

  NOTHROW_DEFAULT_AND_MOVES(Sys<Cons>);
  NOTHROW_DEFAULT_AND_MOVES(Sys<Gens>);

  // Basic status
  Topol topol;
  dim_type dim;
  Status status;
  // The DD pair.
  Sys<Cons> cs;
  Sys<Gens> gs;
  // Saturation info
  Sat sat_c;
  Sat sat_g;
  // Pending stuff
  Sys<Cons> cs_pending;
  Sys<Gens> gs_pending;

  // These methods are meant to be not trivially accessible from Poly.
  // They are anyway accessible using method Poly::impl().
  bool marked_empty() const { return status == Status::EMPTY; }
  bool marked_min() const { return status == Status::MINIMIZED; }
  bool has_pending() const { return status == Status::PENDING; }
  bool has_cs_pending() const {
    return has_pending() && !cs_pending.empty();
  }
  bool has_gs_pending() const {
    return has_pending() && !gs_pending.empty();
  }
  bool has_valid_cons() const { return !has_gs_pending(); }
  bool has_valid_gens() const { return !has_cs_pending(); }

  void set_status(Status s) { status = s; }

  void clear_all();
  void reinit_with_gens(dim_type, Topol, const Sys<Gens>&) = delete;
  void reinit_with_gens(dim_type d, Topol t, Sys<Gens>&& gens);
  void minimize() const; // Only logically const.
  void ensure_valid_cons() const {
    if (!has_valid_cons())
      minimize();
    assert(check_inv(false));
  }
  void ensure_valid_gens() const {
    if (!has_valid_gens())
      minimize();
    assert(check_inv(false));
  }

  dim_type num_equals() const {
    assert(marked_empty() || marked_min());
    return cs.sing_rows.size();
  }
  dim_type num_lines() const {
    assert(marked_empty() || marked_min());
    return gs.sing_rows.size();
  }

  dim_type num_min_cons() const;
  dim_type num_min_gens() const;

  std::tuple<const Sys<Cons>*, const Sys<Cons>*>
  get_sys_ptrs(const Sys<Cons>*) const {
    return std::make_tuple(&cs, &cs_pending);
  }
  std::tuple<const Sys<Gens>*, const Sys<Gens>*>
  get_sys_ptrs(const Sys<Gens>*) const {
    return std::make_tuple(&gs, &gs_pending);
  }

  template <typename Sys, typename Out, typename Pred>
  static Out
  materialize(const Sys& sys, const Sys& pending, Out out,
              bool cons_and_closed, Pred pred);

  // These methods are meant to be accessible from Poly.
  Poly_Impl(const Poly_Impl& y) = default;
  Poly_Impl(Poly_Impl&& y) = default;
  Poly_Impl& operator=(const Poly_Impl& y) = default;
  Poly_Impl& operator=(Poly_Impl&& y) = default;
  ~Poly_Impl() = default;

  explicit
  Poly_Impl(dim_type d = 0,
            Spec_Elem s = Spec_Elem::UNIVERSE,
            Topol t = Topol::CLOSED);

  /* Predicates */

  bool check_inv(bool do_equals = true) const;
  bool is_necessarily_closed() const { return topol == Topol::CLOSED; }

  bool is_empty() const;
  bool is_minimized() const { return !has_pending(); }
  bool is_universe() const;
  bool is_topologically_closed() const;
  bool is_bounded() const;

  bool constrains(Var var) const;
  bool equals(const Poly_Impl& y) const;
  bool contains(const Poly_Impl& y) const;
  bool strictly_contains(const Poly_Impl& y) const {
    const auto& x = *this;
    return x.contains(y) && !y.contains(x);
  }
  bool is_disjoint_from(const Poly_Impl& y) const;

  bool bounds_expr(bool from_below, const Linear_Expr& expr) const;

  /* Queries */

  Topol topology() const { return topol; }
  dim_type space_dim() const { return dim; }
  dim_type affine_dim() const;
  Poly_Con_Rel relation_with(const Con& c) const;
  Poly_Gen_Rel relation_with(const Gen& g) const;
  using Cons_Proxy = Mater_Sys<Sys<Cons>, Poly_Impl>;
  Cons_Proxy cons() const {
    ensure_valid_cons();
    return Cons_Proxy(this);
  }

  using Gens_Proxy = Mater_Sys<Sys<Gens>, Poly_Impl>;
  Gens_Proxy gens() const {
    ensure_valid_gens();
    return Gens_Proxy(this);
  }

  Cons copy_cons() const {
    auto proxy = cons();
    return Cons(proxy.begin(), proxy.end());
  }
  Gens copy_gens() const {
    auto proxy = gens();
    return Gens(proxy.begin(), proxy.end());
  }

  bool min_max(const bool minimizing,
               const Linear_Expr& expr, const Integer& inhomo,
               Integer& num, Integer& den, bool& included, Gen& g) const;
  bool min(const Linear_Expr& expr, const Integer& inhomo,
           Integer& num, Integer& den, bool& included, Gen& g) const {
    return min_max(true, expr, inhomo, num, den, included, g);
  }
  bool max(const Linear_Expr& expr, const Integer& inhomo,
           Integer& num, Integer& den, bool& included, Gen& g) const {
    return min_max(false, expr, inhomo, num, den, included, g);
  }

  /* Modifiers */

  void m_swap(Poly_Impl& y) noexcept {
    using std::swap;
    swap(topol, y.topol);
    swap(dim, y.dim);
    swap(status, y.status);
    swap(cs, y.cs);
    swap(gs, y.gs);
    swap(sat_c, y.sat_c);
    swap(sat_g, y.sat_g);
    swap(cs_pending, y.cs_pending);
    swap(gs_pending, y.gs_pending);
  }

  void set_empty();
  void set_universe();
  void set_topology(Topol t);

  template <typename Iter>
  void add_cons(Iter first, Iter last);
  void add_con(Con c);
  void add_cons(Cons cs) {
    add_cons(std::make_move_iterator(cs.begin()),
             std::make_move_iterator(cs.end()));
  }

  template <typename Iter>
  void add_gens(Iter first, Iter last);
  void add_gen(Gen g);
  void add_gens(Gens gs) {
    add_gens(std::make_move_iterator(gs.begin()),
             std::make_move_iterator(gs.end()));
  }

  void topological_closure_assign();

  template <typename Iter>
  void unconstrain(Iter first, Iter last);
  void unconstrain(Var var) {
    auto d = var.id();
    unconstrain(&d, &d + 1);
  }
  void unconstrain(const Index_Set& vars) {
    unconstrain(vars.begin(), vars.end());
  }

  void intersection_assign(const Poly_Impl& y);
  void poly_hull_assign(const Poly_Impl& y);
  void poly_difference_assign(const Poly_Impl& y);

  void affine_image(Var var, const Linear_Expr& expr,
                    const Integer& inhomo, const Integer& den);
  void affine_preimage(Var var, const Linear_Expr& expr,
                       const Integer& inhomo, const Integer& den);
  void
  parallel_affine_image(const std::vector<Var>& vars,
                        std::vector<Linear_Expr>& exprs,
                        Integers& inhomos, Integers& dens);

  // void
  // generalized_affine_image(Var var,
  //                          Relation_Symbol relsym,
  //                          const Linear_Expr& expr,
  //                          const Integer& inhomo = Integer::zero(),
  //                          const Integer& den = Integer::one());
  // void
  // generalized_affine_preimage(Var var,
  //                             Relation_Symbol relsym,
  //                             const Linear_Expr& expr,
  //                             const Integer& inhomo = Integer::zero(),
  //                             const Integer& den = Integer::one());

  // void generalized_affine_image(const Linear_Expr& lhs,
  //                               Relation_Symbol relsym,
  //                               const Linear_Expr& rhs);
  // void generalized_affine_preimage(const Linear_Expr& lhs,
  //                                  Relation_Symbol relsym,
  //                                  const Linear_Expr& rhs);

  // When the widen_impl pointer is null, h79_widen will be used.
  using widen_type = void (*)(Poly_Impl&, const Poly_Impl&);
  static PPLITE_TLS widen_type widen_impl;
  static widen_type get_widen_impl();
  static void set_widen_impl(widen_type w_impl);

  void widening_assign(const Poly_Impl& y, widen_type widen,
                       const Cons* upto_ptr);

  void time_elapse_assign(const Poly_Impl& y);
  void positive_time_elapse_assign(const Poly_Impl& y);

  // bool simplify_using_context_assign(const Poly_Impl& y);

  // void drop_some_non_integer_points();
  // void drop_some_non_integer_points(const Vars_Set& vars);

  /* Change of space dim */

  void add_space_dims(dim_type m, bool project = false);
  void remove_higher_space_dims(dim_type new_dim);
  template <typename Iter>
  void remove_space_dims(Iter first, Iter last);
  void remove_space_dims(const Index_Set& vars) {
    remove_space_dims(vars.begin(), vars.end());
  }
  void remove_space_dim(Var var) {
    auto dim = var.id();
    remove_space_dims(&dim, &dim + 1);
  }

  void map_space_dims(const Dims& pfunc);
  void concatenate_assign(const Poly_Impl& y);
  void expand_space_dim(Var var, dim_type m) {
    // FIXME.
    (void) var; (void) m;
    assert(false); abort();
  }
  void fold_space_dims(const Index_Set& vars, Var dest);

  Poly_Impl split(const Con& c, Topol t);

  /* Input-output */
  bool ascii_load(std::istream& s);
  void ascii_dump(std::ostream& s) const;

}; // struct Poly_Impl

inline void swap(Poly_Impl& x, Poly_Impl& y) noexcept { x.m_swap(y); }


/* The interface class: mild encapsulation, easy to overcome using impl(). */
class Poly : private Poly_Impl {
public:
  explicit
  Poly(dim_type d = 0,
       Spec_Elem s = Spec_Elem::UNIVERSE,
       Topol t = Topol::CLOSED)
    : Poly_Impl(d, s, t) {}

  // Allow for all possible orderings of the three arguments.
  Poly(dim_type d, Topol t, Spec_Elem s = Spec_Elem::UNIVERSE)
    : Poly_Impl(d, s, t) {}
  Poly(Spec_Elem s, dim_type d, Topol t = Topol::CLOSED)
    : Poly_Impl(d, s, t) {}
  Poly(Spec_Elem s, Topol t, dim_type d)
    : Poly_Impl(d, s, t) {}
  Poly(Topol t, dim_type d, Spec_Elem s = Spec_Elem::UNIVERSE)
    : Poly_Impl(d, s, t) {}
  Poly(Topol t, Spec_Elem s, dim_type d)
    : Poly_Impl(d, s, t) {}

  Poly(const Poly& y) = default;
  Poly(Poly&& y) = default;
  Poly& operator=(const Poly& y) = default;
  Poly& operator=(Poly&& y) = default;
  ~Poly() = default;

  /* Accessors to low level implementation */
  using Impl = Poly_Impl;
  Impl& impl() { return *this; }
  const Impl& impl() const { return *this; }

  /* Types */
  using Impl::Cons_Proxy;
  using Impl::Gens_Proxy;

  /* Predicates */
  using Impl::is_necessarily_closed;
  using Impl::check_inv;
  using Impl::is_empty;
  using Impl::is_universe;
  using Impl::is_minimized;
  using Impl::is_topologically_closed;
  using Impl::is_bounded;
  using Impl::bounds_expr;
  using Impl::constrains;
  bool equals(const Poly& y) const {
    return impl().equals(y.impl());
  }
  bool contains(const Poly& y) const {
    return impl().contains(y.impl());
  }
  bool strictly_contains(const Poly& y) const {
    return impl().strictly_contains(y.impl());
  }
  bool is_disjoint_from(const Poly& y) const {
    return impl().is_disjoint_from(y.impl());
  }

  /* Queries */
  using Impl::topology;
  using Impl::space_dim;
  using Impl::affine_dim;
  using Impl::relation_with;
  using Impl::min_max;
  using Impl::min;
  using Impl::max;
  using Impl::cons;
  using Impl::gens;
  using Impl::copy_cons;
  using Impl::copy_gens;
  using Impl::num_min_cons;
  using Impl::num_min_gens;

  /* Modifiers */
  void m_swap(Poly& y) noexcept { impl().m_swap(y.impl()); }

  using Impl::set_empty;
  using Impl::set_universe;
  using Impl::set_topology;
  using Impl::add_con;
  using Impl::add_cons;
  template <typename Iter>
  void add_cons(Iter first, Iter last) { impl().add_cons(first, last); }

  using Impl::add_gen;
  using Impl::add_gens;
  template <typename Iter>
  void add_gens(Iter first, Iter last) { impl().add_gens(first, last); }

  using Impl::topological_closure_assign;

  template <typename Iter>
  void unconstrain(Iter first, Iter last) {
    impl().unconstrain(first, last);
  }
  using Impl::unconstrain;

  void intersection_assign(const Poly& y) {
    impl().intersection_assign(y.impl());
  }
  void poly_hull_assign(const Poly& y) {
    impl().poly_hull_assign(y.impl());
  }
  void poly_difference_assign(const Poly& y) {
    impl().poly_difference_assign(y.impl());
  }

  void affine_image(Var var, const Linear_Expr& expr,
                    const Integer& inhomo = Integer::zero(),
                    const Integer& den = Integer::one()) {
    impl().affine_image(var, expr, inhomo, den);
  }

  void affine_preimage(Var var, const Linear_Expr& expr,
                       const Integer& inhomo = Integer::zero(),
                       const Integer& den = Integer::one()) {
    impl().affine_preimage(var, expr, inhomo, den);
  }
  using Impl::parallel_affine_image;

  // using Impl::generalized_affine_image;
  // using Impl::generalized_affine_preimage;

  using Impl::widen_type;
  using Impl::get_widen_impl;
  using Impl::set_widen_impl;
  void widening_assign(const Poly& y,
                       widen_type widen = get_widen_impl()) {
    return impl().widening_assign(y.impl(), widen, nullptr);
  }
  void widening_assign(const Poly& y, const Cons& upto_cons,
                       widen_type widen = get_widen_impl()) {
    return impl().widening_assign(y.impl(), widen, &upto_cons);
  }

  void time_elapse_assign(const Poly& y) {
    return impl().time_elapse_assign(y.impl());
  }
  void positive_time_elapse_assign(const Poly& y) {
    return impl().positive_time_elapse_assign(y.impl());
  }

  // bool simplify_using_context_assign(const Poly& y);
  // using Impl::drop_some_non_integer_points;

  /* Change of space dim */

  template <typename Iter>
  void remove_space_dims(Iter first, Iter last) {
    impl().remove_space_dims(first, last);
  }
  void concatenate_assign(const Poly& y) {
    impl().concatenate_assign(y.impl());
  }

  Poly split(const Con& c, Topol t) {
    Poly res;
    res.impl() = impl().split(c, t);
    return res;
  }
  Poly split(const Con& c) { return split(c, topology()); }

  using Impl::add_space_dims;
  using Impl::map_space_dims;
  using Impl::remove_space_dim;
  using Impl::remove_space_dims;
  using Impl::remove_higher_space_dims;
  using Impl::expand_space_dim;
  using Impl::fold_space_dims;

  // This is semantically const,
  // but may affect syntactic representation.
  using Impl::minimize;

  /* Input-output */
  using Impl::ascii_load;
  using Impl::ascii_dump;

}; // class Poly

NOTHROW_MOVES(Poly_Impl);
NOTHROW_MOVES(Poly);

inline void swap(Poly& x, Poly& y) noexcept { x.m_swap(y); }

inline bool operator==(const Poly& x, const Poly& y) { return x.equals(y); }
inline bool operator!=(const Poly& x, const Poly& y) { return !(x == y); }

namespace IO_Operators {

std::ostream& operator<<(std::ostream& s, Topol topol);
std::ostream& operator<<(std::ostream& s, Poly_Impl::Status status);
std::ostream& operator<<(std::ostream& s, const Poly_Impl& ph);
std::ostream& operator<<(std::ostream& s, const Poly& ph);

} // namespace IO_Operators

} // namespace pplite

#endif // !defined(pplite_Poly_hh)
