/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_Poly_Stats_hh
#define pplite_Poly_Stats_hh 1

#include "globals.hh"
#include "clock.hh"
#include "Poly.hh"

namespace pplite {

// An enum providing a distinct value for each abstract operator.
// Note: listed in alphabetical order to avoid confusion;
// Z_OTHER_OPS stand for those operations that either cost nothing
// or are almost never called.
enum class AbsOp {
  ADD_CONS,
  ADD_GENS,
  AFFINE_IMAGES,
  AFFINE_IMAGES_PARALLEL,
  CONSTRAINS,
  CONTAINS,
  COPY_CONS,
  COPY_GENS,
  DIMS_ADD,
  DIMS_MAP,
  DIMS_REM,
  EQUALS,
  IS_EMPTY,
  MINIMIZE,
  MINMAX,
  POLY_HULL,
  RELATION_WITH,
  SPLIT,
  TOPCLOSURE,
  UNCONSTRAIN,
  WIDENING,
  WIDENING_UPTO,
  Z_OTHER_OPS,
  NUM_OPS
}; // enum class AbsOp

constexpr auto absop_size = static_cast<unsigned>(AbsOp::NUM_OPS);

// The array of abstract operator names (indexed by AbsOp).
// Careful: these must match the ordering of AbsOp enumeration values.
constexpr const char* absop_names[absop_size] = {
  "add_cons", "add_gens", "affine_images", "affine_images_parallel",
  "constrains", "contains", "copy_cons", "copy_gens",
  "dims_add", "dims_map", "dims_rem", "equals",
  "is_empty", "minimize", "minmax", "poly_hull",
  "relation_with", "split", "topclosure", "unconstrain",
  "widening", "widening_upto", "other_ops"
};

// A data structure collecting stats for each AbsOp.
struct AbsOp_Stats {
  // Whether or not the destructor should output statistics info.
  bool noisy_dtor_;
  // The collected statistics.
  Time_Stats time_stats[absop_size];

  explicit AbsOp_Stats(bool noisy_dtor = false) : noisy_dtor_(noisy_dtor) {}

  // Clears all statistics.
  void reset();

  // Adds `secs' seconds and `usecs' microseconds to stats for `op',
  // increasing by 1 the number of calls.
  void incr(AbsOp op, time_t secs, time_t usecs) {
    time_stats[static_cast<unsigned>(op)].incr(secs, usecs);
  }

  void dump_op(std::ostream& os, AbsOp op, bool dump_times = false) const;
  void dump(std::ostream& os, bool dump_times = false) const;
  void dump_overall_time(std::ostream& os) const;
  ~AbsOp_Stats();
};

extern AbsOp_Stats absop_stats;

class AbsOp_Clock {
public:
  AbsOp_Clock(AbsOp op, bool noisy = false)
    : clock(), op_(op), noisy_(noisy) {}
  // Not copyable, only moveable.
  AbsOp_Clock(const AbsOp_Clock&) = delete;
  AbsOp_Clock& operator=(const AbsOp_Clock&) = delete;
  AbsOp_Clock(AbsOp_Clock&&) = default;
  AbsOp_Clock& operator=(AbsOp_Clock&&) = default;
  ~AbsOp_Clock() { stop_clock(); }

  void stop_clock() const {
    auto elapsed = clock.elapsed_time();
    if (noisy_) {
      std::cerr << absop_names[static_cast<unsigned>(op_)] << " ";
      Clock::print(std::cerr, elapsed.first, elapsed.second);
      std::cerr << std::endl;
    }
    absop_stats.incr(op_, elapsed.first, elapsed.second);
  }

private:
  Clock clock;
  const AbsOp op_;
  const bool noisy_;
}; // class AbsOp_Clock

template <typename PH>
struct Stats : private PH {
  using Base = PH;

  using typename Base::Impl;

  using typename Base::widen_type;
  using Base::get_widen_impl;
  using Base::set_widen_impl;
  using Base::check_inv;
  using Base::ascii_load;

  // Inherits the constructors: tracking accurate time stats
  // for these is quite tricky (and not really worth the effort).
  using Base::Base;

  Base& base() { return static_cast<Base&>(*this); }
  const Base& base() const { return static_cast<const Base&>(*this); }

  Impl& impl() { return this->Base::impl(); }
  const Impl& impl() const { return this->Base::impl(); }

  void add_con(Con c) {
    AbsOp_Clock clock(AbsOp::ADD_CONS);
    this->Base::add_con(std::move(c));
  }
  template <typename Iter>
  void add_cons(Iter first, Iter last) {
    AbsOp_Clock clock(AbsOp::ADD_CONS);
    this->Base::add_cons(first, last);
  }
  void add_cons(Cons cs) {
    AbsOp_Clock clock(AbsOp::ADD_CONS);
    this->Base::add_cons(std::move(cs));
  }
  void add_gen(Gen g) {
    AbsOp_Clock clock(AbsOp::ADD_GENS);
    this->Base::add_gen(std::move(g));
  }
  template <typename Iter>
  void add_gens(Iter first, Iter last) {
    AbsOp_Clock clock(AbsOp::ADD_GENS);
    this->Base::add_gens(first, last);
  }
  void add_gens(Gens gs) {
    AbsOp_Clock clock(AbsOp::ADD_GENS);
    this->Base::add_gens(std::move(gs));
  }
  void affine_image(Var var,
                    const Linear_Expr& expr,
                    const Integer& inhomo = Integer::zero(),
                    const Integer& den = Integer::one()) {
    AbsOp_Clock clock(AbsOp::AFFINE_IMAGES);
    return this->Base::affine_image(var, expr, inhomo, den);
  }
  void affine_preimage(Var var,
                       const Linear_Expr& expr,
                       const Integer& inhomo = Integer::zero(),
                       const Integer& den = Integer::one()) {
    AbsOp_Clock clock(AbsOp::AFFINE_IMAGES);
    return this->Base::affine_preimage(var, expr, inhomo, den);
  }
  void parallel_affine_image(const std::vector<Var>& vars,
                             std::vector<Linear_Expr>& exprs,
                             std::vector<Integer>& inhomos,
                             std::vector<Integer>& dens) {
    AbsOp_Clock clock(AbsOp::AFFINE_IMAGES_PARALLEL);
    return this->Base::parallel_affine_image(vars, exprs, inhomos, dens);
  }
  bool constrains(Var v) const {
    AbsOp_Clock clock(AbsOp::CONSTRAINS);
    return this->Base::constrains(v);
  }
  bool contains(const Stats& y) const {
    AbsOp_Clock clock(AbsOp::CONTAINS);
    return this->Base::contains(y);
  }
  bool strictly_contains(const Stats& y) const {
    AbsOp_Clock clock(AbsOp::CONTAINS);
    return this->Base::strictly_contains(y);
  }
  Poly::Cons_Proxy cons() const {
    return this->Base::cons();
  }
  Poly::Gens_Proxy gens() const {
    return this->Base::gens();
  }
  Cons copy_cons() const {
    AbsOp_Clock clock(AbsOp::COPY_CONS);
    return this->Base::copy_cons();
  }
  Gens copy_gens() const {
    AbsOp_Clock clock(AbsOp::COPY_GENS);
    return this->Base::copy_gens();
  }
  template <typename Partial_Func>
  void map_space_dims(const Partial_Func& pfunc) {
    AbsOp_Clock clock(AbsOp::DIMS_MAP);
    this->Base::map_space_dims(pfunc);
  }
  void add_space_dims(dim_type m, bool project = false) {
    AbsOp_Clock clock(AbsOp::DIMS_ADD);
    this->Base::add_space_dims(m, project);
  }
  void remove_higher_space_dims(dim_type new_dim) {
    AbsOp_Clock clock(AbsOp::DIMS_REM);
    this->Base::remove_higher_space_dims(new_dim);
  }
  void remove_space_dim(Var v) {
    AbsOp_Clock clock(AbsOp::DIMS_REM);
    this->Base::remove_space_dim(v);
  }
  template <typename Iter>
  void remove_space_dims(Iter first, Iter last) {
    AbsOp_Clock clock(AbsOp::DIMS_REM);
    this->Base::remove_space_dims(first, last);
  }
  void remove_space_dims(const Index_Set& vs) {
    AbsOp_Clock clock(AbsOp::DIMS_REM);
    this->Base::remove_space_dims(vs);
  }
  bool equals(const Stats& y) const {
    AbsOp_Clock clock(AbsOp::EQUALS);
    return this->Base::equals(y);
  }
  bool is_empty() const {
    AbsOp_Clock clock(AbsOp::IS_EMPTY);
    return this->Base::is_empty();
  }
  void minimize() const {
    AbsOp_Clock clock(AbsOp::MINIMIZE);
    return this->Base::minimize();
  }
  bool min_max(const bool minimizing,
               const Linear_Expr& expr, const Integer& inhomo,
               Integer& num, Integer& den, bool& included, Gen& g) const {
    AbsOp_Clock clock(AbsOp::MINMAX);
    return this->Base::min_max(minimizing, expr, inhomo,
                               num, den, included, g);
  }
  bool min(const Linear_Expr& expr, const Integer& inhomo,
           Integer& num, Integer& den, bool& included, Gen& g) const {
    AbsOp_Clock clock(AbsOp::MINMAX);
    return this->Base::min(expr, inhomo, num, den, included, g);
  }
  bool max(const Linear_Expr& expr, const Integer& inhomo,
           Integer& num, Integer& den, bool& included, Gen& g) const {
    AbsOp_Clock clock(AbsOp::MINMAX);
    return this->Base::max(expr, inhomo, num, den, included, g);
  }
  void poly_hull_assign(const Stats& y) {
    AbsOp_Clock clock(AbsOp::POLY_HULL);
    this->Base::poly_hull_assign(y);
  }
  Poly_Con_Rel relation_with(const Con& c) const {
    AbsOp_Clock clock(AbsOp::RELATION_WITH);
    return this->Base::relation_with(c);
  }
  Poly_Gen_Rel relation_with(const Gen& g) const {
    AbsOp_Clock clock(AbsOp::RELATION_WITH);
    return this->Base::relation_with(g);
  }
  void topological_closure_assign() {
    AbsOp_Clock clock(AbsOp::TOPCLOSURE);
    this->Base::topological_closure_assign();
  }
  void unconstrain(Var v) {
    AbsOp_Clock clock(AbsOp::UNCONSTRAIN);
    return this->Base::unconstrain(v);
  }
  void unconstrain(const Index_Set& vs) {
    AbsOp_Clock clock(AbsOp::UNCONSTRAIN);
    return this->Base::unconstrain(vs);
  }
  void widening_assign(const Stats& y,
                       widen_type widen = get_widen_impl()) {
    AbsOp_Clock clock(AbsOp::WIDENING);
    return this->Base::widening_assign(y, widen);
  }
  void widening_assign(const Poly& y, const Cons& upto_cons,
                       widen_type widen = get_widen_impl()) {
    AbsOp_Clock clock(AbsOp::WIDENING_UPTO);
    return this->Base::widening_assign(y, upto_cons, widen);
  }
  // Other ops.
  dim_type affine_dim() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::affine_dim();
  }
  void concatenate_assign(const Stats& y) {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::concatenate_assign(y);
  }
  void expand_space_dim(Var v, dim_type n) {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::expand_space_dim(v, n);
  }
  void fold_space_dims(const Index_Set& vs, Var v) {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::fold_space_dims(vs, v);
  }
  void m_swap(Stats& y) {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::m_swap(y);
  }
  void set_empty() {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::set_empty();
  }
  Stats split(const Con& c, Topol t) {
    AbsOp_Clock clock(AbsOp::SPLIT);
    Stats res;
    res.base() = this->Base::split(c, t);
    return res;
  }
  Stats split(const Con& c) { return split(c, topology()); }
  void intersection_assign(const Stats& y) {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    this->Base::intersection_assign(y);
  }
  bool is_bounded() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::is_bounded();
  }
  bool is_disjoint_from(const Stats& y) const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::is_disjoint_from(y);
  }
  bool is_topologically_closed() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::is_topologically_closed();
  }
  bool is_necessarily_closed() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::is_necessarily_closed();
  }
  bool is_universe() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::is_universe();
  }
  dim_type num_min_cons() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::num_min_cons();
  }
  dim_type num_min_gens() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::num_min_gens();
  }
  void poly_difference_assign(const Stats& y) {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    this->Base::poly_difference_assign(y);
  }
  dim_type space_dim() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::space_dim();
  }
  Topol topology() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::topology();
  }
  void set_topology(Topol t) {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::set_topology(t);
  }
  void time_elapse_assign(const Stats& y) {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    this->Base::time_elapse_assign(y);
  }
  void positive_time_elapse_assign(const Stats& y) {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    this->Base::positive_time_elapse_assign(y);
  }
  size_t total_memory_in_bytes() const {
    AbsOp_Clock clock(AbsOp::Z_OTHER_OPS);
    return this->Base::total_memory_in_bytes();
  }
  void ascii_dump(std::ostream& os) const {
    this->Base::ascii_dump(os);
  }
}; // struct Stats

template <typename PH>
bool operator==(const Stats<PH>& x, const Stats<PH>& y) {
  AbsOp_Clock clock(AbsOp::EQUALS);
  const PH& xx = x.base();
  const PH& yy = y.base();
  return xx == yy;
}

template <typename PH>
void swap(Stats<PH>& x, Stats<PH>& y) { x.m_swap(y); }

template <typename PH>
inline bool
operator!=(const Stats<PH>& x, const Stats<PH>& y) {
  return !(x == y);
}

namespace IO_Operators {

template <typename PH>
std::ostream& operator<<(std::ostream& s, const Stats<PH>& ph) {
  return s << ph.base();
}

} // namespace IO_Operators

using Poly_Stats = Stats<Poly>;
NOTHROW_MOVES(Poly_Stats);

} // namespace pplite

#endif // !defined(pplite_Poly_Stats_hh)
