/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_U_Poly_hh
#define pplite_U_Poly_hh 1

#include "globals.hh"
#include "Poly.hh"
#include "Poly_min.hh"
#include "Poly_templ.hh"

#include <cassert>
#include <iostream>
#include <iterator>
#include <numeric>
#include <vector>

namespace pplite {

template <typename PH>
class U_Wrap {
public:
  using Kernel = PH;

  explicit
  U_Wrap(dim_type d = 0, Spec_Elem s = Spec_Elem::UNIVERSE,
         Topol t = Topol::CLOSED)
    : info(d, not_a_dim()), kernel(0, s, t) { assert(check_inv()); }
  U_Wrap(dim_type d, Topol t, Spec_Elem s = Spec_Elem::UNIVERSE)
    : U_Wrap(d, s, t) {}
  U_Wrap(Spec_Elem s, dim_type d, Topol t = Topol::CLOSED)
    : U_Wrap(d, s, t) {}
  U_Wrap(Spec_Elem s, Topol t, dim_type d)
    : U_Wrap(d, s, t) {}
  U_Wrap(Topol t, dim_type d, Spec_Elem s = Spec_Elem::UNIVERSE)
    : U_Wrap(d, s, t) {}
  U_Wrap(Topol t, Spec_Elem s, dim_type d)
    : U_Wrap(d, s, t) {}

  explicit U_Wrap(const Kernel& ph)
    : U_Wrap(ph.space_dim(), ph.topology(),
             ph.is_empty() ? Spec_Elem::EMPTY : Spec_Elem::UNIVERSE) {
    if (is_empty())
      return;
    const auto& cs = ph.cons();
    add_shell_cons(cs.begin(), cs.end());
  }

  // Implicit conversion to Kernel.
  operator Kernel() const {
    Kernel res = Kernel(space_dim(), topology(),
                        is_empty() ? Spec_Elem::EMPTY : Spec_Elem::UNIVERSE);
    if (!is_empty())
      res.add_cons(copy_cons());
    return res;
  }

  U_Wrap(const U_Wrap& y) = default;
  U_Wrap& operator=(const U_Wrap& y) = default;
  U_Wrap(U_Wrap&& y) = default;
  U_Wrap& operator=(U_Wrap&& y) = default;
  ~U_Wrap() = default;

  /* Predicates */
  bool check_inv() const;
  bool is_necessarily_closed() const { return kernel.is_necessarily_closed(); }
  bool is_empty() const { return kernel.is_empty(); }
  bool is_minimized() const { return kernel.is_minimized(); }
  bool is_universe() const {
    return !is_empty() && kernel.space_dim() == 0;
  }
  bool is_topologically_closed() const {
    return kernel.is_topologically_closed();
  }
  bool is_bounded() const {
    if (is_empty())
      return true;
    return space_dim() == kernel.space_dim() && kernel.is_bounded();
  }
  // FIXME: bounds_expr
  bool constrains(Var var) const {
    return is_empty() || is_kernel_dim(var.id());
  }
  bool equals(const U_Wrap& y) const;
  bool contains(const U_Wrap& y) const;
  bool strictly_contains(const U_Wrap& y) const;
  bool is_disjoint_from(const U_Wrap& y) const;

  /* Queries */
  Topol topology() const { return kernel.topology(); }
  dim_type space_dim() const { return info.size(); }
  dim_type affine_dim() const {
    if (is_empty()) return 0;
    return space_dim() - kernel.space_dim() + kernel.affine_dim();
  }
  Poly_Con_Rel relation_with(const Con& c) const;
  Poly_Gen_Rel relation_with(const Gen& g) const;
  bool min_max(const bool minimizing,
               const Linear_Expr& expr, const Integer& inhomo,
               Integer& num, Integer& den, bool& included, Gen& g) const;
  bool min(const Linear_Expr& expr, const Integer& inhomo,
           Integer& num, Integer& den, bool& included, Gen& g) const {
    return min_max(true, expr, inhomo, num, den, included, g);
  }
  bool max(const Linear_Expr& expr, const Integer& inhomo,
           Integer& num, Integer& den, bool& included, Gen& g) const {
    return min_max(false, expr, inhomo, num, den, included, g);
  }
  // FIXME: provide appropriate proxy for cons() and gens().
  Cons cons() const { return copy_cons(); }
  Gens gens() const { return copy_gens(); }
  Cons copy_cons() const;
  Gens copy_gens() const;
  dim_type num_min_cons() const { return kernel.num_min_cons(); }
  dim_type num_min_gens() const {
    if (is_empty())
      return 0;
    return kernel.num_min_gens()
      // Add a line for each unconstrained dim.
      + std::count(info.begin(), info.end(), not_a_dim());
  }

  /* Modifiers */

  void m_swap(U_Wrap& y) noexcept {
    using std::swap;
    swap(info, y.info);
    kernel.m_swap(y.kernel);
  }

  void set_empty() {
    info.assign(info.size(), not_a_dim());
    set_kernel(Spec_Elem::EMPTY);
  }
  void set_universe() {
    info.assign(info.size(), not_a_dim());
    set_kernel(Spec_Elem::UNIVERSE);
  }
  void set_topology(Topol t) { kernel.set_topology(t); }

  void add_con(Con c) { add_shell_cons(&c, &c + 1); }
  // Note: no point in passing move_iterators.
  void add_cons(Cons cs) { add_shell_cons(cs.begin(), cs.end()); }
  template <typename Iter>
  void add_cons(Iter first, Iter last) { add_shell_cons(first, last); }

  void add_gen(Gen g) { add_shell_gens(&g, &g + 1); }
  // Note: no point in passing move_iterators.
  void add_gens(Gens gs) { add_shell_gens(gs.begin(), gs.end()); }
  template <typename Iter>
  void add_gens(Iter first, Iter last) { add_shell_gens(first, last); }

  void topological_closure_assign() {
    kernel.topological_closure_assign();
  }

  template <typename Iter>
  void unconstrain(Iter first, Iter last) {
    if (first == last)
      return;
    if (is_empty())
      return;
    unconstrain_kernel(first, last);
    check_kernel_for_unconstrained();
    assert(check_inv());
  }
  void unconstrain(Var var) {
    auto d = var.id();
    unconstrain(&d, &d + 1);
  }
  void unconstrain(const Index_Set& vars) {
    unconstrain(vars.begin(), vars.end());
  }

  void intersection_assign(const U_Wrap& y);
  void poly_hull_assign(const U_Wrap& y);
  void poly_difference_assign(const U_Wrap&) {
    assert(false && "Unimplemented");
    abort();
  }

  void affine_image(Var var, const Linear_Expr& expr,
                    const Integer& inhomo = Integer::zero(),
                    const Integer& den = Integer::one()) {
    affine_image_aux(false, var, expr, inhomo, den);
  }
  void affine_preimage(Var var, const Linear_Expr& expr,
                       const Integer& inhomo = Integer::zero(),
                       const Integer& den = Integer::one()) {
    affine_image_aux(true, var, expr, inhomo, den);
  }
  void
  parallel_affine_image(std::vector<Var>& vars,
                        std::vector<Linear_Expr>& exprs,
                        Integers& inhomos, Integers& dens);

  using widen_type = typename Kernel::widen_type;
  static widen_type get_widen_impl() {
    return Kernel::get_widen_impl();
  }
  static void set_widen_impl(widen_type widen) {
    Kernel::set_widen_impl(widen);
  }

  void widening_assign(const U_Wrap& y, widen_type widen, const Cons* upto_ptr);
  void widening_assign(const U_Wrap& y,
                       widen_type widen = get_widen_impl()) {
    widening_assign(y, widen, nullptr);
  }
  void widening_assign(const U_Wrap& y, const Cons& upto_cons,
                       widen_type widen = get_widen_impl()) {
    widening_assign(y, widen, &upto_cons);
  }

  void time_elapse_assign(const U_Wrap& y);
  void positive_time_elapse_assign(const U_Wrap& y);

  /* Change of space dim */
  void add_space_dims(dim_type m, bool project = false);
  void map_space_dims(const Dims& pfunc);
  void remove_higher_space_dims(dim_type new_dim) {
    assert(new_dim <= space_dim());
    Dims vars(space_dim() - new_dim);
    std::iota(vars.begin(), vars.end(), new_dim);
    remove_space_dims(vars.begin(), vars.end());
  }
  template <typename Iter>
  void remove_space_dims(Iter first, Iter last) {
    unconstrain(first, last);
    erase_using_sorted_indices(info, first, last);
    assert(check_inv());
  }
  void remove_space_dims(const Index_Set& vars) {
    remove_space_dims(vars.begin(), vars.end());
  }
  void remove_space_dim(Var var) {
    auto dim = var.id();
    remove_space_dims(&dim, &dim + 1);
  }
  void concatenate_assign(const U_Wrap& y);
  void expand_space_dim(Var var, dim_type m);
  void fold_space_dims(const Index_Set& vars, Var dest);

  // Semantically const, but may affect syntactic representation.
  void minimize() const { kernel.minimize(); }

  /* Input-output */
  bool ascii_load(std::istream& s);
  void ascii_dump(std::ostream& s) const;

private:
  Dims info;
  Kernel kernel;

  Dims get_inverse_info() const;
  Cons kernel_2_shell_cons() const;
  Gens kernel_2_shell_gens() const;

  Con kernel_2_shell_con(const Con& k_c, const Dims& inverse) const;
  Gen kernel_2_shell_gen(const Gen& k_g, const Dims& inverse) const;

  bool is_kernel_dim(dim_type shell_dim) const {
    return info[shell_dim] != not_a_dim();
  }
  bool is_unconstrained_dim(dim_type shell_dim) const {
    return !is_kernel_dim(shell_dim);
  }

  dim_type& kernel_dim(dim_type shell_dim) {
    assert(is_kernel_dim(shell_dim));
    return info[shell_dim];
  }
  dim_type kernel_dim(dim_type shell_dim) const {
    assert(is_kernel_dim(shell_dim));
    return info[shell_dim];
  }

  bool all_kernel_dims(const Linear_Expr& expr) const {
    for (auto i = expr.space_dim(); i-- > 0; ) {
      if (is_unconstrained_dim(i) && !expr.get(i).is_zero())
        return false;
    }
    return true;
  }

  Linear_Expr shell_2_kernel_expr(const Linear_Expr& s_expr) const;

  Con shell_2_kernel_con(const Con& s_c) const;
  Gen shell_2_kernel_gen(const Gen& s_g) const;

  void decr_info_helper(dim_type k_dim) {
    for (auto& kd : info) {
      if (kd != not_a_dim() && kd > k_dim)
        --kd;
    }
  }

  void set_kernel(Spec_Elem kind) {
    kernel = Kernel(0, kind, kernel.topology());
  }

  template <typename Iter>
  void add_shell_cons(Iter first, Iter last) {
    if (first == last)
      return;
    // First pass: update info with dims that will enter the kernel.
    const dim_type old_k_dim = kernel.space_dim();
    dim_type new_k_dims = 0;
    for (Iter i = first; i != last; ++i) {
      const auto& c = *i;
      for (dim_type j = 0, j_end = c.space_dim(); j != j_end; ++j) {
        if (is_kernel_dim(j) || c.coeff(Var(j)).is_zero())
          continue;
        info[j] = old_k_dim + new_k_dims;
        ++new_k_dims;
      }
    }
    kernel.add_space_dims(new_k_dims);

    // Second pass: translate shell cons into kernel cons.
    Cons kernel_cs;
    for (Iter i = first; i != last; ++i) {
      const auto& s_c = *i;
      Con k_c = shell_2_kernel_con(s_c);
      if (!k_c.linear_expr().is_zero())
        kernel_cs.push_back(std::move(k_c));
      else if (k_c.is_inconsistent()) {
        set_empty();
        return;
      }
    }
    kernel.add_cons(std::move(kernel_cs));
    kernel.minimize();
    if (kernel.is_empty())
      set_empty();
  }

  template <typename Iter>
  void add_shell_gens(const Iter first, const Iter last) {
    if (first == last)
      return;
    if (is_universe())
      return;
    if (is_empty()) {
      std::iota(info.begin(), info.end(), 0);
      kernel.add_space_dims(space_dim());
      kernel.add_gens(first, last);
      check_kernel_for_unconstrained();
      assert(check_inv());
      return;
    }

    bool nothing_todo = true;
    for (Iter i = first; nothing_todo && i != last; ++i) {
      const auto& sg = *i;
      if (sg.is_point() || sg.is_closure_point()) {
        nothing_todo = false;
        break;
      }
      for (dim_type j = 0, j_end = sg.space_dim(); j != j_end; ++j) {
        if (is_kernel_dim(j) && sg.coeff(Var(j)) != 0) {
          nothing_todo = false;
          break;
        }
      }
    }
    if (nothing_todo)
      return;

    Gens kernel_gs;
    for (Iter i = first; i != last; ++i) {
      const auto& s_g = *i;
      Linear_Expr k_expr;
      for (dim_type j = 0, j_end = s_g.space_dim(); j != j_end; ++j) {
        if (is_kernel_dim(j) && s_g.coeff(Var(j)) != 0)
          k_expr.set(Var(kernel_dim(j)), s_g.coeff(Var(j)));
      }
      if ((s_g.is_ray() || s_g.is_line()) && k_expr.is_zero())
        continue;
      Gen k_g(s_g.type(), std::move(k_expr), s_g.impl().inhomo);
      kernel_gs.push_back(std::move(k_g));
    }
    kernel.add_gens(std::move(kernel_gs));
    kernel.minimize();
    check_kernel_for_unconstrained();
    assert(check_inv());
  }

  template <typename Iter>
  void unconstrain_kernel(Iter first, Iter last) {
    assert(first != last);
    assert(!is_empty());
    Index_Set k_dims;
    for ( ; first != last; ++first) {
      dim_type sd = *first;
      assert(sd < space_dim());
      if (is_kernel_dim(sd)) {
        k_dims.set(kernel_dim(sd));
        info[sd] = not_a_dim();
      }
    }
    // Remove the kernel variables (all at once).
    kernel.remove_space_dims(k_dims);
    // Fix info to match the removed kernel variables (one at a time).
    dim_type found = 0;
    for (auto kd : k_dims) {
      decr_info_helper(kd - found);
      ++found;
    }
    // Note: NOT checking if other kernel dims became unconstrained.
    // This is up to the caller (class invariant may be broken here).
  }
  void unconstrain_kernel(const Index_Set& vars) {
    unconstrain_kernel(vars.begin(), vars.end());
  }

  void sync_kernel_dims(const U_Wrap& y) const;
  void check_kernel_for_unconstrained() const;

  void affine_image_aux(bool preimage, Var var, const Linear_Expr& expr,
                        const Integer& inhomo, const Integer& den);

}; // class U_Wrap

namespace IO_Operators {

template <typename PH>
std::ostream& operator<<(std::ostream& s, const U_Wrap<PH>& ph);

} // namespace IO_Operators

template <typename PH>
inline void
swap(U_Wrap<PH>& x, U_Wrap<PH>& y) noexcept { x.m_swap(y); }

template <typename PH>
inline bool
operator==(const U_Wrap<PH>& x, const U_Wrap<PH>& y) { return x.equals(y); }
template <typename PH>
inline bool
operator!=(const U_Wrap<PH>& x, const U_Wrap<PH>& y) { return !(x == y); }

// Explicit template instantiation (declaration).
extern template class U_Wrap<Poly>;
using U_Poly = U_Wrap<Poly>;

} // namespace pplite

#endif // !defined(pplite_U_Poly_hh)
