/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_mater_iterator_hh
#define pplite_mater_iterator_hh 1

#include "globals.hh"
#include "Con.hh"
#include "Gen.hh"
#include "support_utils.hh"

#include <memory>
#include <iterator>
#include <cassert>

namespace pplite {

template <typename> class Mater_Iter;

template <typename Sys, typename Impl>
struct Mater_Sys {
  // Types.
  using sys_type = Sys;
  using container_type = typename sys_type::rows_type;
  using size_type = typename container_type::size_type;
  using value_type = typename container_type::value_type;
  using cache_type = std::vector<std::unique_ptr<value_type>>;
  using const_iterator = Mater_Iter<Mater_Sys>;

  // Data members.
  const sys_type* ptr1;
  const sys_type* ptr2;
  size_type sk_offset;
  size_type ns_offset;
  size_type end_pos;
  mutable cache_type cache;
  const Impl* impl;

  // Constructor.
  Mater_Sys(const Impl* p_impl)
    : impl(p_impl) {
    std::tie(ptr1, ptr2) = impl->get_sys_ptrs(ptr1);
    sk_offset = ptr1->sing_rows.size() + ptr2->sing_rows.size();
    ns_offset = sk_offset + ptr1->sk_rows.size() + ptr2->sk_rows.size();
    end_pos = ns_offset + ptr1->ns_rows.size() + ptr2->ns_rows.size();
    cache.resize(end_pos - ns_offset);
  }

  // Movable.
  Mater_Sys(Mater_Sys&&) = default;
  Mater_Sys& operator=(Mater_Sys&&) = default;
  ~Mater_Sys() = default;
  // Non-copiable.
  Mater_Sys() = delete;
  Mater_Sys(const Mater_Sys&) = delete;
  Mater_Sys& operator=(const Mater_Sys&) = delete;

  static bool is_skippable(const Con* c) {
    return c->is_strict_inequality() && c->is_tautological();
  }
  static bool is_skippable(const Gen*) { return false; }

  bool is_skippable(size_type pos) const {
    assert(pos < end_pos);
    // Generators are never skippable.
    if (std::is_same<value_type, Gen>::value)
      return false;
    // Equations are not skippable.
    if (pos < sk_offset)
      return false;
    if (pos < ns_offset)
      return is_skippable(get_value_ptr(pos));
    if (impl->is_necessarily_closed())
      return true;
    // Non-skel element: we only skip the efc in ptr1
    // (the non-pending constraints).
    if (impl->sat_g.num_rows() == 0)
      return false;
    pos -= ns_offset;
    return pos < ptr1->ns_rows.size() &&
      detail::is_empty_face_cutter(ptr1->ns_rows[pos],
                                   impl->sat_g, impl->gs.sk_rows);
  }

  void maybe_cache(size_type cache_pos,
                   const sys_type* ptr,
                   size_type ns_pos) const {
    if (cache[cache_pos] == nullptr) {
      auto m = detail::materialize(ptr->ns_rows[ns_pos], ptr->sk_rows);
      cache[cache_pos].reset(new value_type(std::move(m)));
    }
  }

  const value_type* get_value_ptr(size_type pos) const {
    if (pos < sk_offset) {
      // Singular.
      if (pos < ptr1->sing_rows.size())
        return &ptr1->sing_rows[pos];
      pos -= ptr1->sing_rows.size();
      return &ptr2->sing_rows[pos];
    }
    if (pos < ns_offset) {
      // Skeleton.
      pos -= sk_offset;
      if (pos < ptr1->sk_rows.size())
        return &ptr1->sk_rows[pos];
      pos -= ptr1->sk_rows.size();
      return &ptr2->sk_rows[pos];
    }
    // Non-skeleton.
    assert(pos < end_pos);
    pos -= ns_offset;
    if (pos < ptr1->ns_rows.size()) {
      maybe_cache(pos, ptr1, pos);
      return cache[pos].get();
    }
    maybe_cache(pos, ptr2, pos - ptr1->ns_rows.size());
    return cache[pos].get();
  }

  const_iterator cbegin() const { return const_iterator(this, true); }
  const_iterator cend() const { return const_iterator(this, false); }
  const_iterator begin() const { return cbegin(); }
  const_iterator end() const { return cend(); }

  size_type size() const { return std::distance(begin(), end()); }

}; // Mater_Sys

template <typename Mater_Sys>
class Mater_Iter {
public:
  using mater_sys_type = Mater_Sys;
  using size_type = typename mater_sys_type::size_type;
  const mater_sys_type* ptr;
  size_type pos;

  using base_iter = typename mater_sys_type::container_type::const_iterator;
  using base_traits = std::iterator_traits<base_iter>;

  using iterator_category = std::forward_iterator_tag;
  using value_type = typename base_traits::value_type;
  using pointer = typename base_traits::pointer;
  using reference = typename base_traits::reference;
  using difference_type = typename base_traits::difference_type;

  Mater_Iter(const mater_sys_type* msys, bool start)
    : ptr(msys), pos(0) {
    assert(ptr != nullptr);
    if (start)
      skip_forward();
    else
      pos = ptr->end_pos;
  }

  Mater_Iter() = default;
  Mater_Iter(const Mater_Iter&) = default;
  Mater_Iter(Mater_Iter&&) = default;
  Mater_Iter& operator=(const Mater_Iter&) = default;
  Mater_Iter& operator=(Mater_Iter&&) = default;
  ~Mater_Iter() = default;

  void skip_forward() {
    while (pos < ptr->end_pos && ptr->is_skippable(pos))
      ++pos;
  }

  reference operator*() const { return *(ptr->get_value_ptr(pos)); }
  pointer operator->() const { return ptr->get_value_ptr(pos); }

  Mater_Iter& operator++() {
    assert(pos != ptr->end_pos);
    ++pos;
    skip_forward();
    return *this;
  }
  Mater_Iter operator++(int) {
    auto old = *this;
    operator++();
    return old;
  }

}; // class Mater_Iter

template <typename Mater_Sys>
inline bool
operator==(const Mater_Iter<Mater_Sys>& x, const Mater_Iter<Mater_Sys>& y) {
  assert(x.ptr == y.ptr);
  return x.pos == y.pos;
}

template <typename Mater_Sys>
inline bool
operator!=(const Mater_Iter<Mater_Sys>& x, const Mater_Iter<Mater_Sys>& y) {
  return !(x == y);
}

} // namespace pplite

#endif // !defined(pplite_mater_iterator_hh)
