/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite_test.hh"

bool
test01() {
  Var A(0);
  Var B(1);

  Poly ph(2, Topol::NNC);
  Cons cs;
  cs.push_back(A > 0);
  cs.push_back(B >= 0);
  cs.push_back(B <= 1);
  ph.add_cons(cs);
  ph.minimize();

  cs = ph.copy_cons();
  auto strict_before = std::count_if(cs.begin(), cs.end(),
                                     std::mem_fn(&Con::is_strict_inequality));
  print_cons(ph, "=== ph before adding closure point ===");

  ph.add_gen(closure_point(2*B));
  ph.minimize();

  cs = ph.copy_cons();
  auto strict_after = std::count_if(cs.begin(), cs.end(),
                                    std::mem_fn(&Con::is_strict_inequality));
  print_cons(ph, "=== ph after adding closure point ===");

  Poly kr(2, Topol::NNC);
  cs.clear();
  cs.push_back(A > 0);
  cs.push_back(B >= 0);
  cs.push_back(B < 2);
  kr.add_cons(cs);
  kr.minimize();
  print_cons(kr, "=== expected ===");

  return (strict_after == (strict_before + 1)) && (ph == kr);
}

bool
test02() {
  Var A(0);
  Var B(1);

  Poly ph(2, Topol::NNC);
  Cons cs;
  cs.push_back(A > 1);
  cs.push_back(B >= 1);
  cs.push_back(B <= 3);
  ph.add_cons(cs);
  ph.minimize();

  ph.add_gen(closure_point());
  ph.minimize();

  cs = ph.copy_cons();
  int ph_strict = std::count_if(cs.begin(), cs.end(),
                                std::mem_fn(&Con::is_strict_inequality));
  Poly kr(2, Topol::NNC);
  cs.clear();
  cs.push_back(3*A - B > 0);
  cs.push_back(B > 0);
  cs.push_back(B <= 3);
  kr.add_cons(cs);
  kr.minimize();

  cs = kr.copy_cons();
  int kr_strict = std::count_if(cs.begin(), cs.end(),
                                std::mem_fn(&Con::is_strict_inequality));
  return  kr_strict == ph_strict;
}

bool
test04() {
  Var A(0);
  Var B(1);
  Var C(2);

  Poly ph(3, Topol::NNC, Spec_Elem::EMPTY);
  ph.add_gen(point());
  ph.add_gen(point(A));
  ph.add_gen(point(B));
  ph.add_gen(point(A+B));
  ph.add_gen(ray(B));
  ph.minimize();

  ph.add_gen(line(B));
  ph.add_gen(ray(A));
  ph.add_gen(point(-A));

  ph.add_con(A>0);
  ph.add_con(B<=1);
  ph.add_gen(ray(C));
  ph.add_gen(point(2*B));

  ph.add_con(A+B+C <= 5);
  ph.add_gen(point(6*B));
  ph.minimize();

  return ph.check_inv();
}

bool
test05() {
  Var A(0);
  Var B(1);
  Var C(2);

  Poly ph(3, Topol::NNC, Spec_Elem::EMPTY);
  ph.add_gen(point());
  ph.add_gen(point(A));
  ph.add_gen(point(B));
  ph.add_gen(point(A+B));
  ph.add_gen(point(C));
  ph.add_gen(point(C+A));
  ph.add_gen(point(C+B));
  ph.add_gen(point(C+A+B));
  ph.minimize();

  ph.add_gen(ray(A+B));
  ph.add_gen(ray(A));
  ph.add_gen(ray(B));
  ph.add_gen(ray(C));
  ph.minimize();

  ph.add_gen(point(-A));
  ph.add_gen(point(-B));
  ph.add_gen(point(-C));
  ph.add_gen(point(-B-C));
  ph.add_gen(point(-A-B));
  ph.add_gen(point(-A-C));
  ph.add_gen(point(-A-B-C));
  ph.minimize();

  return ph.check_inv();
}

/* Note:
 * The commented stats are to be checked in the opt build
 * since in the debug one many more conversions may be called.
 */
bool
test06() {
  Var A(0);
  Var B(1);
  Var C(2);
  Var D(3);

  Poly ph(4, Topol::NNC, Spec_Elem::EMPTY);
  ph.add_gen(point());
  ph.add_gen(point(A));
  ph.add_gen(point(B));
  ph.minimize();

  // point + efc
  ph.add_gen(point(A+B));
  ph.add_gen(point(2*A + 2*B));
  ph.add_gen(point(2*A));
  ph.add_gen(point(2*B));
  ph.minimize();  // 4/4 - was 4/4

  // ray + efc
  ph.add_gen(ray(B));
  // point + rfc
  ph.add_gen(point(A - B));
  // ph.minimize(); // 2/2 - was 0/2

  // ray + rfc
  ph.add_gen(ray(A));
  // ph.minimize(); // 0/1 - was 0/1

  // ray + positivity
  ph.add_gen(ray(-A + B));
  // ph.minimize(); // 1/1 - was 0/1

  // point + positivity
  ph.add_gen(point(C));
  // ph.minimize(); // 1/1 - was 0/1

  // point + rfc
  ph.add_gen(point(A-B+C));
  // ph.minimize(); // 0/1 - was 0/1

  ph.minimize();  // tot: 4/5 - was 0/5

  // line + rfc
  ph.add_gen(line(C));
  ph.minimize();  // 1/1 - was 0/1

  return ph.check_inv();
}

BEGIN_MAIN
  DO_TEST(test01);
  DO_TEST(test02);
/* test03 moved into Poly_Impl */
  DO_TEST(test04);
  DO_TEST(test05);
  DO_TEST(test06);
END_MAIN
